from nhpylm.npylm cimport NPYLM, SHPYLMNode, STables, CHPYLMNode, CTables
from nhpylm.sequence cimport Sequence, BOU, EOU
from nhpylm.random_utils cimport random_choice, bernoulli
import numpy as np
cimport numpy as np
from libc.math cimport log
np.import_array()
DTYPE = np.float64

cdef void apply_hyperparameters_learning(NPYLM npylm, list train_sequences, 
                                        bint d_theta_learning, bint poisson_learning):
    """
    Learn 
        - theta of all depths for both SHPYLM, CHPYLM
        - probability of all possible segment lengths
        - lambda value for poisson correction
    
    Parameters
    ----------
    npylm : NPYLM
        npylm wich we want to update and learn hyperparameters 
    train_sequences : list of Sequences
        list of training sequences represented as Sequences, mainly for the poisson lambda calculation
    d_theta_learning : bint
        boolean whether we want to include the d and theta learning
    poisson_learning : bint
        boolean whether we want to include the poisson lambda and len k prob
    """
    if poisson_learning:
        update_poisson_lambda(npylm, train_sequences)
        update_poisson_k_probs(npylm)
    if d_theta_learning:
        update_shpylm_d_theta(npylm)
        update_chpylm_d_theta(npylm)

cdef void update_poisson_lambda(NPYLM npylm, list train_sequences):
    """
    Update the lambda value for poisson correction using the Gamma distribution.
    new_lambda = Gamma(a, b), where 
        a = Sum_{segments}(len(segment)*t_wh) of root
        b = (init_poisson_b + t_h) of root 
    
    Parameters
    ----------
    npylm : NPYLM
        npylm wich we want to update and learn hyperparameters 
    train_sequences : list of Sequences
        list of training sequences represented as Sequences
    """
    cdef float a = npylm.init_poisson_a
    cdef float b = npylm.init_poisson_b + npylm.shpylm_root.t_h
    cdef str segment
    cdef int segment_length
    cdef int tables_count
    cdef Sequence sequence
    cdef set used_segments = set()
    cdef SHPYLMNode shpylm_root = npylm.shpylm_root
    cdef STables segment_restaurant

    # Get a, b from all possible unique segments
    for sequence in train_sequences:
        for segment in sequence.segmentation:
            segment_length = len(segment)
            if (segment_length <= npylm.max_segment_size) and (not segment in used_segments):
                used_segments.add(segment)
                segment_restaurant = shpylm_root.tables[segment]
                tables_count = len(segment_restaurant.tables)
                a += (tables_count * segment_length)

    # Sample new lambda from a,b
    npylm.poisson_lambda = np.random.gamma(a, 1/b)


cdef void update_poisson_k_probs(NPYLM npylm, int segment_samples = 20000):
    """
    Sample some number of random segments (by default 20000) regarding the distribution of
    CHPYLM (for each new segment sample characters till you get EOU). Keep the statistics about their lengths. 
    Then update poisson_k_probs of npylm regarding the new statistics.
    
    Parameters
    ----------
    npylm : NPYLM
        npylm wich we want to update and learn hyperparameters 
    segment_samples : int
        number of segments that are randomly generated
    """
    cdef np.ndarray length_occurences = np.zeros([npylm.max_segment_size], dtype=DTYPE)
    cdef int i
    cdef list vocabulary = list(npylm.character_vocabulary)
    cdef list context
    cdef list next_character_probs
    cdef str character
    cdef int character_index
    cdef int final_segment_length
    cdef CHPYLMNode chpylm_root = npylm.chpylm_root

    for _ in range(segment_samples):
        # sample random segment from CHPYLM
        context = [BOU]
        next_character_probs = []
        # find the first not-eos character
        for character in vocabulary:
            next_character_probs.append(chpylm_root.get_pwh_probability(character, context, 1.0/len(vocabulary), 1.0))
        character = vocabulary[random_choice(next_character_probs)]
        context.append(character)
        # sample next characters till you find the EOU character
        while character != EOU and len(context) < npylm.max_segment_size + 1: # first element is BOU
            next_character_probs = []
            for character in vocabulary + [EOU]:
                next_character_probs.append(chpylm_root.get_pwh_probability(character, context, 1.0/len(vocabulary), 1.0))
            character_index = random_choice(next_character_probs)
            # In case of EOU, don't add character to context
            if character_index == len(vocabulary):
                character = EOU
            else:
                character = vocabulary[character_index]
                context.append(character)
        # store the final k length into statistics
        final_segment_length = len(context) - 1
        length_occurences[final_segment_length-1] += 1



    for i in range(npylm.max_segment_size):
        # Apply Laplace smoothing
        npylm.poisson_k_probs[i] = (length_occurences[i] + 1)/(segment_samples + npylm.max_segment_size)
       




# XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
# XXX          Segment Hierarchical Pitman-Yor Tree Hyperparameters          XXX
# XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX


cdef void update_shpylm_d_theta(NPYLM npylm):
    """
    Update d and theta arrays of SHPYLM in NPYLM.
    First recursivaly precompute sum(1-y_ui), sum(1-z_uwkj), sum(y_ui), sum(log(x_u))
    over all nodes for each depth separately.
    After that compute new 'd' using Beta priors as Beta(d_a+sum(1-y_ui), d_b+sum(1-z_uwkj))
        where d_a and d_b are defautly set hyperparameters
    And compute new 'theta' using Gamma priors as Gamma(theta_alpha+sum(u_i), 1/(theta_beta+sum(log(x_u))))
        where theta_alpha and theta_beta are defautly set hyperparameters
    for all shpylm tree depths.

    Parameters
    ----------
    npylm : NPYLM
        npylm wich we want to update and learn hyperparameters 
    """
    cdef int i
    cdef float alpha, beta
    # Init value arrays for d,theta updates
    cdef np.ndarray sum1_minus_y_ui = np.zeros([npylm.shpylm_max_depth+1], dtype=DTYPE)
    cdef np.ndarray sum1_minus_z_uwkj = np.zeros([npylm.shpylm_max_depth+1], dtype=DTYPE)
    cdef np.ndarray sumy_ui = np.zeros([npylm.shpylm_max_depth+1], dtype=DTYPE)
    cdef np.ndarray sumlogx_u = np.zeros([npylm.shpylm_max_depth+1], dtype=DTYPE)
    # Set the first depth
    sum1_minus_y_ui[0] = __get_shpylm_1_minus_y_ui(npylm.shpylm_root)
    sum1_minus_z_uwkj[0] = __get_shpylm_1_minus_z_uwkj_sum(npylm.shpylm_root)
    sumy_ui[0] = __get_shpylm_y_ui_sum(npylm.shpylm_root)
    sumlogx_u[0] = __get_shpylm_logx_u(npylm.shpylm_root)
    # Fill array values for all depths recursivaly
    __recursive_shpylm_d_theta_preparation(npylm.shpylm_root, sum1_minus_y_ui, sum1_minus_z_uwkj, sumy_ui, sumlogx_u)
    
    # Set new d, theta for all depths
    for i in range(npylm.shpylm_max_depth+1):
        # sample d
        alpha = npylm.d_a + sum1_minus_y_ui[i]
        beta = npylm.d_b + sum1_minus_z_uwkj[i]
        npylm.shpylm_ds[i] = np.random.beta(alpha, beta)
        # sample theta
        shape = npylm.theta_alpha + sumy_ui[i]
        scale = 1.0/(npylm.theta_beta - sumlogx_u[i])
        npylm.shpylm_thetas[i] = np.random.gamma(shape, scale)


cdef void __recursive_shpylm_d_theta_preparation(SHPYLMNode node, np.ndarray sum1_minus_y_ui, 
                                                np.ndarray sum1_minus_z_uwkj, np.ndarray sumy_ui,
                                                np.ndarray sumlogx_u):
    """
    Recursivaly precompute sum(1-y_ui), sum(1-z_uwkj), sum(y_ui), sum(log(x_u))
    over all nodes for each depth separately.

    Parameters
    ----------
    node : SHPYLMNode
        SHPYLM node of tree we want to add to our precomputed values
    sum1_minus_y_ui : np.ndarray
        array of floats of sum(1-y_ui) for all depths
    sum1_minus_z_uwkj : np.ndarray
        array of floats of sum(1-z_uwkj) for all depths
    sumy_ui : np.ndarray
        array of floats of sum(y_ui) for all depths
    sumlogx_u : np.ndarray
        array of floats of sum(log(x_u)) for all depths
    """
    cdef SHPYLMNode child
    cdef str context
    for context in node.children:
        child = node.children[context]
        # Update value arrays
        sum1_minus_y_ui[child.depth] += __get_shpylm_1_minus_y_ui(child)
        sum1_minus_z_uwkj[child.depth] += __get_shpylm_1_minus_z_uwkj_sum(child)
        sumy_ui[child.depth] += __get_shpylm_y_ui_sum(child)
        sumlogx_u[child.depth] += __get_shpylm_logx_u(child)
        # Recursive call
        __recursive_shpylm_d_theta_preparation(child, sum1_minus_y_ui, sum1_minus_z_uwkj, sumy_ui, sumlogx_u)


cdef float __get_shpylm_1_minus_y_ui(SHPYLMNode node):
    """
    Sample sum_{i=1..t_h}(1-bernoulli((theta)/(theta+d*i))) over all tables in the specific node.
    
    Parameters
    ----------
    node : SHPYLMNode
        SHPYLM node of tree we want to add to our precomputed values
    Returns
    -------
    final_sum : float
        sum_{i=1..t_h}(1-bernoulli((theta)/(theta+d*i)))
    """
    cdef float final_sum = 0.0
    cdef float prob
    cdef int i
    cdef NPYLM npylm = node.npylm
    if node.t_h >= 2:
        for i in range(1,node.t_h):
            prob = npylm.shpylm_thetas[node.depth] / \
                    (npylm.shpylm_thetas[node.depth] + npylm.shpylm_ds[node.depth] * i)
            final_sum += (1-bernoulli(prob))
    return final_sum

cdef float __get_shpylm_1_minus_z_uwkj_sum(SHPYLMNode node):
    """
    Sample sum_{all segments}(sum_{all tables}(sum_{j=1..c_whk}(1-bernoulli((j-1)/(j-d)))) for the specific node
    where c_whk is count of customer at the specific table.
    
    Parameters
    ----------
    node : SHPYLMNode
        SHPYLM node of tree we want to add to our precomputed values
    Returns
    -------
    final_sum : float
        sum_{all segments}(sum_{all tables}(sum_{j=1..c_whk}(1-bernoulli((j-1)/(j-d))))
    """
    cdef float final_sum = 0.0
    cdef str segment
    cdef int c_whk
    cdef float prob
    cdef int j
    cdef NPYLM npylm = node.npylm
    cdef STables restaurant
    for segment in node.tables:
        restaurant = node.tables[segment]
        for c_whk in restaurant.tables:
            if c_whk >= 2:
                for j in range(1,c_whk):
                    prob = (j - 1) / (j - npylm.shpylm_ds[node.depth])
                    final_sum += (1 - bernoulli(prob))
    return final_sum

cdef float __get_shpylm_y_ui_sum(SHPYLMNode node):
    """
    Sample sum_{i=1..t_h}(bernoulli((theta)/(theta+d*i))) over all tables in the specific node.
    
    Parameters
    ----------
    node : SHPYLMNode
        SHPYLM node of tree we want to add to our precomputed values
    Returns
    -------
    final_sum : float
        sum_{i=1..t_h}(bernoulli((theta)/(theta+d*i)))
    """
    cdef float final_sum = 0.0
    cdef float prob
    cdef int i
    cdef NPYLM npylm = node.npylm
    if node.t_h >= 2:
        for i in range(1,node.t_h):
            prob = npylm.shpylm_thetas[node.depth] / \
                    (npylm.shpylm_thetas[node.depth] + npylm.shpylm_ds[node.depth] * i)
            final_sum += bernoulli(prob)
    return final_sum

cdef float __get_shpylm_logx_u(SHPYLMNode node):
    """
    Sample log(Beta(theta+1,c_h-1)) for the specific node.
    
    Parameters
    ----------
    node : SHPYLMNode
        SHPYLM node of tree we want to add to our precomputed values
    Returns
    -------
    logx_u : float
        log(Beta(theta+1,c_h-1)) for t_h >= 2, otherwise 0.0
    """
    cdef NPYLM npylm = node.npylm
    if node.t_h >= 2:
        return log(np.random.beta(npylm.shpylm_thetas[node.depth] + 1, node.c_h - 1))
    else:
        return 0.0



# XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
# XXX         Character Hierarchical Pitman-Yor Tree Hyperparameters        XXX
# XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX




cdef void update_chpylm_d_theta(NPYLM npylm):
    """
    Update d and theta arrays of CHPYLM in NPYLM.
    First recursivaly precompute sum(1-y_ui), sum(1-z_uwkj), sum(y_ui), sum(log(x_u))
    over all nodes for each depth separately.
    After that compute new 'd' using Beta priors as Beta(d_a+sum(1-y_ui), d_b+sum(1-z_uwkj))
        where d_a and d_b are defautly set hyperparameters
    And compute new 'theta' using Gamma priors as Gamma(theta_alpha+sum(u_i), 1/(theta_beta+sum(log(x_u))))
        where theta_alpha and theta_beta are defautly set hyperparameters
    for all shpylm tree depths.

    Parameters
    ----------
    npylm : NPYLM
        npylm wich we want to update and learn hyperparameters 
    """
    cdef int i
    cdef float alpha, beta
    # Init value arrays for d,theta updates
    cdef np.ndarray sum1_minus_y_ui = np.zeros([npylm.chpylm_max_depth+1], dtype=DTYPE)
    cdef np.ndarray sum1_minus_z_uwkj = np.zeros([npylm.chpylm_max_depth+1], dtype=DTYPE)
    cdef np.ndarray sumy_ui = np.zeros([npylm.chpylm_max_depth+1], dtype=DTYPE)
    cdef np.ndarray sumlogx_u = np.zeros([npylm.chpylm_max_depth+1], dtype=DTYPE)
    # Set the first depth
    sum1_minus_y_ui[0] = __get_chpylm_1_minus_y_ui(npylm.chpylm_root)
    sum1_minus_z_uwkj[0] = __get_chpylm_1_minus_z_uwkj_sum(npylm.chpylm_root)
    sumy_ui[0] = __get_chpylm_y_ui_sum(npylm.chpylm_root)
    sumlogx_u[0] = __get_chpylm_logx_u(npylm.chpylm_root)
    # Fill array values for all depths recursivaly
    __recursive_chpylm_d_theta_preparation(npylm.chpylm_root, sum1_minus_y_ui, sum1_minus_z_uwkj, sumy_ui, sumlogx_u)
    
    # Set new d, theta for all depths
    for i in range(npylm.chpylm_max_depth+1):
        # sample d
        alpha = npylm.d_a + sum1_minus_y_ui[i]
        beta = npylm.d_b + sum1_minus_z_uwkj[i]
        npylm.chpylm_ds[i] = np.random.beta(alpha, beta)
        # sample theta
        shape = npylm.theta_alpha + sumy_ui[i]
        scale = 1.0/(npylm.theta_beta - sumlogx_u[i])
        npylm.chpylm_thetas[i] = np.random.gamma(shape, scale)


cdef void __recursive_chpylm_d_theta_preparation(CHPYLMNode node, np.ndarray sum1_minus_y_ui, 
                                                np.ndarray sum1_minus_z_uwkj, np.ndarray sumy_ui,
                                                np.ndarray sumlogx_u):
    """
    Recursivaly precompute sum(1-y_ui), sum(1-z_uwkj), sum(y_ui), sum(log(x_u))
    over all nodes for each depth separately.

    Parameters
    ----------
    node : CHPYLMNode
        CHPYLM node of tree we want to add to our precomputed values
    sum1_minus_y_ui : np.ndarray
        array of floats of sum(1-y_ui) for all depths
    sum1_minus_z_uwkj : np.ndarray
        array of floats of sum(1-z_uwkj) for all depths
    sumy_ui : np.ndarray
        array of floats of sum(y_ui) for all depths
    sumlogx_u : np.ndarray
        array of floats of sum(log(x_u)) for all depths
    """
    cdef CHPYLMNode child
    cdef str context
    for context in node.children:
        child = node.children[context]
        # Update arrays
        sum1_minus_y_ui[child.depth] += __get_chpylm_1_minus_y_ui(child)
        sum1_minus_z_uwkj[child.depth] += __get_chpylm_1_minus_z_uwkj_sum(child)
        sumy_ui[child.depth] += __get_chpylm_y_ui_sum(child)
        sumlogx_u[child.depth] += __get_chpylm_logx_u(child)
        # Recursive call
        __recursive_chpylm_d_theta_preparation(child, sum1_minus_y_ui, sum1_minus_z_uwkj, sumy_ui, sumlogx_u)


cdef float __get_chpylm_1_minus_y_ui(CHPYLMNode node):
    """
    Sample sum_{i=1..t_h}(1-bernoulli((theta)/(theta+d*i))) over all tables in the specific node.
    
    Parameters
    ----------
    node : CHPYLMNode
        CHPYLM node of tree we want to add to our precomputed values
    Returns
    -------
    final_sum : float
        sum_{i=1..t_h}(1-bernoulli((theta)/(theta+d*i)))
    """
    cdef float final_sum = 0.0
    cdef float prob
    cdef int i
    cdef NPYLM npylm = node.npylm
    if node.t_h >= 2:
        for i in range(1,node.t_h):
            prob = npylm.chpylm_thetas[node.depth] / \
                    (npylm.chpylm_thetas[node.depth] + npylm.chpylm_ds[node.depth] * i)
            final_sum += (1-bernoulli(prob))
    return final_sum

cdef float __get_chpylm_1_minus_z_uwkj_sum(CHPYLMNode node):
    """
    Sample sum_{all segments}(sum_{all tables}(sum_{j=1..c_whk}(1-bernoulli((j-1)/(j-d)))) for the specific node
    where c_whk is count of customer at the specific table.
    
    Parameters
    ----------
    node : CHPYLMNode
        CHPYLM node of tree we want to add to our precomputed values
    Returns
    -------
    final_sum : float
        sum_{all segments}(sum_{all tables}(sum_{j=1..c_whk}(1-bernoulli((j-1)/(j-d))))
    """
    cdef float final_sum = 0.0
    cdef str segment
    cdef int c_whk
    cdef float prob
    cdef int j
    cdef NPYLM npylm = node.npylm
    cdef CTables restaurant
    for segment in node.tables:
        restaurant = node.tables[segment]
        for c_whk in restaurant.tables:
            if c_whk >= 2:
                for j in range(1,c_whk):
                    prob = (j - 1) / (j - npylm.chpylm_ds[node.depth])
                    final_sum += (1 - bernoulli(prob))
    return final_sum

cdef float __get_chpylm_y_ui_sum(CHPYLMNode node):
    """
    Sample sum_{i=1..t_h}(bernoulli((theta)/(theta+d*i))) over all tables in the specific node.
    
    Parameters
    ----------
    node : CHPYLMNode
        CHPYLM node of tree we want to add to our precomputed values
    Returns
    -------
    final_sum : float
        sum_{i=1..t_h}(bernoulli((theta)/(theta+d*i)))
    """
    cdef float final_sum = 0.0
    cdef float prob
    cdef int i
    cdef NPYLM npylm = node.npylm
    if node.t_h >= 2:
        for i in range(1,node.t_h):
            prob = npylm.chpylm_thetas[node.depth] / \
                    (npylm.chpylm_thetas[node.depth] + npylm.chpylm_ds[node.depth] * i)
            final_sum += bernoulli(prob)
    return final_sum

cdef float __get_chpylm_logx_u(CHPYLMNode node):
    """
    Sample log(Beta(theta+1,c_h-1)) for the specific node.
    
    Parameters
    ----------
    node : CHPYLMNode
        CHPYLM node of tree we want to add to our precomputed values
    Returns
    -------
    logx_u : float
        log(Beta(theta+1,c_h-1)) for t_h >= 2, otherwise 0.0
    """
    cdef NPYLM npylm = node.npylm
    if node.t_h >= 2:
        return log(np.random.beta(npylm.chpylm_thetas[node.depth] + 1, node.c_h - 1))
    else:
        return 0.0


from nhpylm.npylm cimport NPYLM
from nhpylm.sequence cimport Sequence, EOS, BOS
from nhpylm.random_utils cimport random_choice
import numpy as np
cimport numpy as np
np.import_array()
DTYPE = np.float64


cdef void blocked_gibbs_iteration(NPYLM npylm, list sequences):
    """
    Process one blocked gibbs iteration for all training sequences.
    First shuffle sequences randomly. Sample each sequence - remove sequence segments
    from NPYLM, get new segmentation, add new sequence segments to NPYLM.
    At the end of all sequence sampling also sample hyperparameters.

    Parameters
    ----------
    hpylm : NPYLM
        current state of hpylm language model
    sequence : list
        list of current training sequences we are sampling, their segmentation will be changed
    """
    cdef int sequence_id
    cdef np.ndarray rand_indices = np.arange(len(sequences))

    np.random.shuffle(rand_indices)
    for sequence_id in rand_indices:
        npylm.remove_sequence(sequences[sequence_id])
        __optmize_sequence_segmentation(npylm, sequences[sequence_id])
        npylm.add_sequence(sequences[sequence_id])

cdef void __optmize_sequence_segmentation(NPYLM npylm, Sequence sequence):
    """
    Process the forward filtering to get precomputed alpha array.
    With the alpha values, do the backward sampling to create a new segmentation.

    Parameters
    ----------
    npylm : NPYLM
        current state of npylm language model
    sequence : Sequence
        current sequence we are sampling, its segmentation will be changed after this function
    """

    cdef np.ndarray alpha
    cdef dict bigram_cache_p
    alpha, bigram_cache_p = __forward_filtering(npylm, sequence)
    __backward_sampling(npylm, sequence, alpha, bigram_cache_p)


cdef tuple __forward_filtering(NPYLM npylm, Sequence sequence):
    """
    Process the forward filtering and precompute alpha array with marginalized probabilities
    for all t, k (t is the position in sequence, k is the length of the last segment).
    The scaling is used for avoiding of underflowing. The original paper used expsumlog().

    Parameters
    ----------
    npylm : NPYLM
        current state of npylm language model
    sequence : Sequence
        current sequence we are sampling, its segmentation will be changed after this function
    Returns
    -------
    alpha : np.ndarray
        precompute alpha[t,k] array with marginalized probabilities for all t, k 
        (t is the position in sequence, k is the length of the last segment)
    bigram_cache_p : dict of dicts of floats
        dictionary of cached probabilities of bigrams in order to decrease number of 
        get_bigram_probability calls, first key is the second gram, second key is a first gram
    """
    cdef int sequence_len = len(sequence.sequence_string)
    cdef int max_segment_size = npylm.max_segment_size
    cdef np.ndarray alpha = np.zeros([sequence_len+1, max_segment_size+1], dtype=DTYPE)
    cdef dict bigram_cache_p = {} # first dictionary keys are second grams, second - inner - dictionary keys are first grams
    cdef float prob
    cdef np.ndarray scaling_alpha = np.zeros([sequence_len+1], dtype=DTYPE)
    cdef int t, k, j
    cdef float sum_prob
    cdef float prod_scaling
    cdef float sum_alpha_t

    alpha[0, 0] = 1.0
    for t in range(1, sequence_len+1):
        prod_scaling = 1.0
        sum_alpha_t = 0.0
        for k in range(1, min(max_segment_size, t)+1):
            if k != 1:
                prod_scaling *= scaling_alpha[t-k+1]
            sum_prob = 0.0
            if t-k == 0:
                # first gram is an <bos> beggining of sequence
                # Cache probabilities to avoid of still calling get_bigram_probability function
                if sequence.sequence_string[t-k:t] in bigram_cache_p and BOS in bigram_cache_p[sequence.sequence_string[t-k:t]]:
                    prob = bigram_cache_p[sequence.sequence_string[t-k:t]][BOS]
                else:
                    # second gram: (t-k+1):(t) (by the "vector indexing")
                    prob = npylm.get_bigram_probability(BOS, sequence.sequence_string[t-k:t])
                    if (not sequence.sequence_string[t-k:t] in bigram_cache_p) or (not BOS in bigram_cache_p[sequence.sequence_string[t-k:t]]):
                        bigram_cache_p[sequence.sequence_string[t-k:t]] = {}
                    bigram_cache_p[sequence.sequence_string[t-k:t]][BOS] = prob

                sum_prob += (prob * alpha[0, 0])
            else:
                # for j in range(1, t-k) in the original word segmentation paper - we have to consider max_size
                for j in range(1, min(max_segment_size, t-k)+1): 
                    # Cache probabilities to avoid of still calling get_bigram_probability function
                    if sequence.sequence_string[t-k:t] in bigram_cache_p and sequence.sequence_string[t-k-j:t-k] in bigram_cache_p[sequence.sequence_string[t-k:t]]:
                        prob = bigram_cache_p[sequence.sequence_string[t-k:t]][sequence.sequence_string[t-k-j:t-k]]
                    else:
                        # first gram: (t-k-j+1):(t-k), second gram: (t-k+1):(t) (by the "vector indexing")
                        prob = npylm.get_bigram_probability(sequence.sequence_string[t-k-j:t-k], sequence.sequence_string[t-k:t])
                        if (not sequence.sequence_string[t-k:t] in bigram_cache_p) or (not sequence.sequence_string[t-k-j:t-k] in bigram_cache_p[sequence.sequence_string[t-k:t]]):
                            bigram_cache_p[sequence.sequence_string[t-k:t]] = {}
                        bigram_cache_p[sequence.sequence_string[t-k:t]][sequence.sequence_string[t-k-j:t-k]] = prob
                    sum_prob += (prob * alpha[t-k, j])

            alpha[t, k] = sum_prob*prod_scaling
            sum_alpha_t += sum_prob*prod_scaling

        # Perform scaling to avoid underflowing
        for k in range(1, min(max_segment_size, t)+1):
            alpha[t, k] /= sum_alpha_t
        scaling_alpha[t] = 1.0/sum_alpha_t

    return alpha, bigram_cache_p


cdef void __backward_sampling(NPYLM npylm, Sequence sequence, np.ndarray alpha, dict bigram_cache_p):
    """
    Do the backward sampling to get optimized segmentation with np.random.choice of 
    all k candidates k ~ p(w_{i} | c_{t-k+1}^{t}, Theta) * alpha[t][k] in each step.

    Parameters
    ----------
    npylm : NPYLM
        current state of npylm language model
    sequence : Sequence
        current sequence we are sampling, its segmentation will be changed after this function
    alpha : np.ndarray
        precompute alpha[t,k] array with marginalized probabilities for all t, k 
        (t is the position in sequence, k is the length of the last segment).
    bigram_cache_p : dict of dicts of floats
        dictionary of cached probabilities of bigrams in order to decrease number of 
        get_bigram_probability calls, first key is the second gram, second key is a first gram
    """
    cdef int t = len(sequence.sequence_string)
    cdef int max_segment_size = npylm.max_segment_size
    cdef int k
    cdef list probs
    cdef float prob
    cdef list k_candidates
    cdef list borders = [t]
    cdef str w = EOS

    while t > 0:   
        probs = []
        k_candidates = []
        for k in range(1, min(max_segment_size, t)+1):
            k_candidates.append(k)
            # Use bigram prob cache if possible
            if w in bigram_cache_p and sequence.sequence_string[t-k:t] in bigram_cache_p[w]:
                prob = bigram_cache_p[w][sequence.sequence_string[t-k:t]]
            else:
                # first gram: (t-k+1):(k), second gram: w (by the "vector indexing")
                prob = npylm.get_bigram_probability(sequence.sequence_string[t-k:t], w)
            probs.append(prob * alpha[t, k])

        k = k_candidates[random_choice(probs)]
        w = sequence.sequence_string[t-k:t]
        t -= k
        borders.append(t)

    
    borders.reverse()
    sequence.set_segmentation(borders)


from nhpylm.npylm cimport NPYLM
from nhpylm.sequence cimport EOS, BOS
import numpy as np
cimport numpy as np
np.import_array()
DTYPE = np.float64


cdef list viterbi_segment_sequences(NPYLM npylm, list sequences_str):
    """
    Process viterbi segmentation for all testing sequences based on the hpylm model.

    Parameters
    ----------
    npylm : NPYLM
        current state of npylm language model
    sequence : list of strings
        list of current training sequences we are sampling, their segmentation will be changed
    Returns
    -------
    final_segmetation : list of lists of strings
        list of segmented sequences, one sequence segmentation is represented as list of strigns(~segments)
    """
    cdef list final_segmentation = []
    for sequence_string in sequences_str:
        final_segmentation.append(__get_sequence_segmentation(npylm, sequence_string))
    return final_segmentation



cdef list __get_sequence_segmentation(NPYLM npylm, str sequence_str):
    """
    Process the forward filtering to get precomputed alpha array.
    With the alpha values, do the backward sampling to predict the segmentation.

    Parameters
    ----------
    npylm : NPYLM
        current state of npylm language model
    sequence_str : string
        current sequence we are sampling, its segmentation will be changed after this function
    Returns
    -------
    final_segmetation : list of lists of strings
        list of segmented sequences, one sequence segmentation is represented as list of strigns(~segments)
    """
    cdef np.ndarray alpha
    cdef dict bigram_cache_p
    alpha, bigram_cache_p = __forward_filtering(npylm, sequence_str)
    return __get_segments_with_backward_sampling(npylm, sequence_str, alpha, bigram_cache_p)


cdef tuple __forward_filtering(NPYLM npylm, str sequence_str):
    """
    Process the forward filtering and precompute alpha array with marginalized probabilities
    for all t, k (t is the position in sequence, k is the length of the last segment).
    The scaling is used for avoiding of underflowing. The original paper used expsumlog().

    Parameters
    ----------
    npylm : NPYLM
        current state of npylm language model
    sequence_str : string
        current sequence we are sampling, its segmentation will be changed after this function
    Returns
    -------
    alpha : np.ndarray
        precompute alpha[t,k] array with marginalized probabilities for all t, k 
        (t is the position in sequence, k is the length of the last segment)
    bigram_cache_p : dict of dicts of floats
        dictionary of cached probabilities of bigrams in order to decrease number of 
        get_bigram_probability calls, first key is the second gram, second key is a first gram
    """
    cdef int sequence_len = len(sequence_str)
    cdef int max_segment_size = npylm.max_segment_size
    cdef np.ndarray alpha = np.zeros([sequence_len+1, max_segment_size+1], dtype=DTYPE)
    cdef dict bigram_cache_p = {} # first dictionary keys are second grams, second - inner - dictionary keys are first grams
    cdef float prob
    cdef np.ndarray scaling_alpha = np.zeros([sequence_len+1], dtype=DTYPE)
    cdef int t, k, j
    cdef float sum_prob
    cdef float prod_scaling
    cdef float sum_alpha_t

    alpha[0, 0] = 1.0
    for t in range(1, sequence_len+1):
        prod_scaling = 1.0
        sum_alpha_t = 0.0
        for k in range(1, min(max_segment_size, t)+1):
            if k != 1:
                prod_scaling *= scaling_alpha[t-k+1]
            sum_prob = 0.0
            if t-k == 0:
                # first gram is an <bos> beggining of sequence
                # Cache probabilities to avoid of still calling get_bigram_probability function
                if sequence_str[t-k:t] in bigram_cache_p and BOS in bigram_cache_p[sequence_str[t-k:t]]:
                    prob = bigram_cache_p[sequence_str[t-k:t]][BOS]
                else:
                    # second gram: (t-k+1):(t) (by the "vector indexing")
                    prob = npylm.get_bigram_probability(BOS, sequence_str[t-k:t])
                    if (not sequence_str[t-k:t] in bigram_cache_p) or (not BOS in bigram_cache_p[sequence_str[t-k:t]]):
                        bigram_cache_p[sequence_str[t-k:t]] = {}
                    bigram_cache_p[sequence_str[t-k:t]][BOS] = prob

                sum_prob += (prob * alpha[0, 0])
            else:
                # for j in range(1, t-k) in the original word segmentation paper - we have to consider max_size
                for j in range(1, min(max_segment_size, t-k)+1): 
                    # Cache probabilities to avoid of still calling get_bigram_probability function
                    if sequence_str[t-k:t] in bigram_cache_p and sequence_str[t-k-j:t-k] in bigram_cache_p[sequence_str[t-k:t]]:
                        prob = bigram_cache_p[sequence_str[t-k:t]][sequence_str[t-k-j:t-k]]
                    else:
                        # first gram: (t-k-j+1):(t-k), second gram: (t-k+1):(t) (by the "vector indexing")
                        prob = npylm.get_bigram_probability(sequence_str[t-k-j:t-k], sequence_str[t-k:t])
                        if (not sequence_str[t-k:t] in bigram_cache_p) or (not sequence_str[t-k-j:t-k] in bigram_cache_p[sequence_str[t-k:t]]):
                            bigram_cache_p[sequence_str[t-k:t]] = {}
                        bigram_cache_p[sequence_str[t-k:t]][sequence_str[t-k-j:t-k]] = prob
                    sum_prob += (prob * alpha[t-k, j])

            alpha[t, k] = sum_prob*prod_scaling
            sum_alpha_t += sum_prob*prod_scaling

        # Perform scaling to avoid underflowing
        for k in range(1, min(max_segment_size, t)+1):
            alpha[t, k] /= sum_alpha_t
        scaling_alpha[t] = 1.0/sum_alpha_t

    return alpha, bigram_cache_p


cdef list __get_segments_with_backward_sampling(NPYLM npylm, str sequence_str, np.ndarray alpha, dict bigram_cache_p):
    """
    Get segmentation using the backward sampling regarding the precompued alpha.
    The final segmentation is chosen
    based on argmax of k ~ p(w_{i} | c_{t-k+1}^{t}, Theta) * alpha[t][k] in each step.

    Parameters
    ----------
    npylm : NPYLM
        current state of npylm language model
    sequence_str : string
        current sequence we are sampling, its segmentation will be changed after this function
    alpha : np.ndarray
        precompute alpha[t,k] array with marginalized probabilities for all t, k 
        (t is the position in sequence, k is the length of the last segment)
    bigram_cache_p : dict of dicts of floats
        dictionary of cached probabilities of bigrams in order to decrease number of 
        get_bigram_probability calls, first key is the second gram, second key is a first gram
    Returns
    -------
    final_segmetation : list of lists of strings
        list of segmented sequences, one sequence segmentation is represented as list of strigns(~segments)
    """
    cdef int t = len(sequence_str)
    cdef int max_segment_size = npylm.max_segment_size
    cdef list segmented_sequence = []
    cdef int i, j, k
    cdef list probs
    cdef float prob
    cdef list k_candidates
    cdef list borders = [t]
    cdef str w = EOS


    while t > 0:   
        probs = []
        k_candidates = []
        for k in range(1, min(max_segment_size, t)+1):
            k_candidates.append(k)
            # Use bigram prob cache if possible
            if w in bigram_cache_p and sequence_str[t-k:t] in bigram_cache_p[w]:
                prob = bigram_cache_p[w][sequence_str[t-k:t]]
            else:
                # first gram: (t-k+1):(k), second gram: w (by the "vector indexing")
                prob = npylm.get_bigram_probability(sequence_str[t-k:t], w)
            probs.append(prob * alpha[t, k])
        k = k_candidates[np.argmax(probs)]
        w = sequence_str[t-k:t]
        t -= k
        borders.append(t)


    borders.reverse()
    for i, j in zip(borders[:-1], borders[1:]):
        segmented_sequence.append(sequence_str[i:j])

    return segmented_sequence


from libc.stdlib cimport rand, RAND_MAX
from libc.math cimport exp, pow

cdef int random_choice(list probabilities):
    """
    Do the random choice regarding the probabilities distribution.
    Return the random index of coming list corresponding to the list's distribution.
    The probabilities list doesn't have to sum up to 1, the function computes it itself.
    In case that sum of probabilities values sum up to 0, the uniform distribution is used.

    Parameters
    ----------
    probabilities : list of floats
        list of probabilities, doesn't have to sum up to 1
    Returns
    -------
    i : int
        chosen index of the random choice function regarding the probabilities distribution
    """
    cdef int i
    cdef float p
    cdef float random_number = float(rand())/float(RAND_MAX)
    cdef float prob_sum = 0.0
    cdef float left_border = 0.0
    cdef float normalized_p

    for p in probabilities:
        prob_sum += p
    if prob_sum == 0:
        # Uniform distribution
        for i in range(len(probabilities)):
            normalized_p = 1.0/len(probabilities)
            if random_number <= left_border + normalized_p:
                return i
            left_border += normalized_p
    else:
        # Given probabilities distribution
        for i, p in enumerate(probabilities):
            normalized_p = p/prob_sum
            if random_number <= left_border + normalized_p:
                return i
            left_border += normalized_p

cdef float poisson(int k, float lam):
    """
    Poisson distribution. 
    Po(k, lambda) = exp(-lambda)*((lambda^k)/(k!))

    Parameters
    ----------
    k : int
        size k
    lam : float
        lambda
    Returns
    -------
    poisson : float
        value get from the poisson distribution
    """
    cdef int k_factorial
    cdef int i
    if k == 0:
        k_factorial = 1
    else:
        k_factorial = 1
        for i in range(1, k+1):
            k_factorial *= i

    poisson = exp(-lam)*((pow(lam, k))/(k_factorial))

    return poisson

cdef int bernoulli(float prob):
    """
    Bernoulli distribution. 
    Return 1 with probability prob, return 0 with probability 1-prob.

    Parameters
    ----------
    prob : float
        probability of sampling 1
    Returns
    -------
    bernoulli : int
        sampled 1 or 0, depending on the prob
    """
    cdef float random_number = float(rand())/float(RAND_MAX)
    if random_number < prob:
        return 1
    else:
        return 0



from nhpylm.sequence cimport Sequence
import numpy as np
cimport numpy as np

cdef class NPYLM():
    cdef int max_segment_size
    cdef float beta_stops, beta_passes
    cdef int shpylm_max_depth, chpylm_max_depth
    cdef list shpylm_ds, shpylm_thetas, chpylm_ds, chpylm_thetas
    cdef float init_d, init_theta, d_a, d_b, theta_alpha, theta_beta
    cdef float theta
    cdef float poisson_lambda
    cdef float init_poisson_a
    cdef float init_poisson_b
    cdef list poisson_k_probs
    cdef set character_vocabulary
    cdef dict root_tables_context_lengths
    cdef int last_removed_shpylm_table_index
    cdef CHPYLMNode chpylm_root # root of character HPYLM tree
    cdef SHPYLMNode shpylm_root # root of segment HPYLM tree
    cdef void add_sequence(self, Sequence sequence)
    cdef void remove_sequence(self, Sequence sequence)
    cdef float get_bigram_probability(self, str first_gram, str second_gram)
    cdef float get_segmentation_log_probability(self, list sequence_segmentation)
    cdef void add_segment_to_chpylm(self, str segment)
    cdef void remove_segment_from_chpylm(self, str segment)
    cdef float get_G0_probability(self, str segment)



# XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
# XXX          Segment Hierarchical Pitman-Yor Tree Node          XXX
# XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

cdef class SHPYLMNode():
    cdef int depth
    cdef SHPYLMNode parent
    cdef NPYLM npylm
    cdef str hpylm_type # 't' for character hpylm, 's' for shant hpylm
    cdef dict children # dictionary of deeper context HPYLNodes
    cdef str context # one previous element
    cdef dict tables # dict of all tables - key: segment, value: Tables
    cdef dict c_wh # c(w|h) .. count of all customers 
    cdef int c_h # sum of all counts in c(w|h)
    cdef int t_h # t_{h} .. sum of counts of tables in tables dictionary
    cdef bint add_segment(self, str segment, list context, float pwhcomma)
    cdef bint remove_segment(self, str segment, list context)
    cdef float get_pwh_probability(self, str segment, list context, float pwhcomma)


cdef class STables():
    # t_hw could be get as len(self.tables)
    cdef list tables # list of ints - customer counts
    cdef str segment
    cdef SHPYLMNode hpylmnode
    cdef bint add_customer(self, float pwhcomma, int t_h)
    cdef bint remove_customer(self)


# XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
# XXX         Character Hierarchical Pitman-Yor Tree Node        XXX
# XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

cdef class CHPYLMNode():
    cdef CHPYLMNode parent
    cdef NPYLM npylm
    cdef int depth
    cdef int stops # stop counts
    cdef int passes # pass counts
    cdef str hpylm_type # 't' for character hpylm, 's' for shant hpylm
    cdef dict children # dictionary of deeper context HPYLNodes
    cdef str context # one previous element
    cdef dict tables # dict of all tables - key: segment, value: Tables
    cdef dict c_wh # c(w|h) .. count of all customers 
    cdef int c_h # sum of all counts in c(w|h)
    cdef int t_h # t_{h} .. sum of counts of tables in tables dictionary
    cdef bint add_character(self, str character, list context, float pwhcomma)
    cdef bint remove_character(self, str character, list context)
    cdef float get_pwh_probability(self, str character, list context, float pwhcomma, float prev_pass_product)
    cdef int sample_context_length(self, str character, list context, float pwhcomma)
    cdef void __fill_sample_prob_table(self, list sample_prob_table, str character, list context, float pwhcomma, float prev_pass_product)

cdef class CTables():
    # t_hw could be get as len(self.tables)
    cdef list tables # list of ints - customer counts
    cdef str character
    cdef CHPYLMNode hpylmnode
    cdef bint add_customer(self, float pwhcomma, int t_h)
    cdef bint remove_customer(self)


from nhpylm.npylm cimport NPYLM, SHPYLMNode, CHPYLMNode
import numpy as np
cimport numpy as np
cdef void apply_hyperparameters_learning(NPYLM npylm, list train_sequences, bint d_theta_learning, bint poisson_learning)
cdef void update_poisson_lambda(NPYLM npylm, list train_sequences)
cdef void update_poisson_k_probs(NPYLM npylm, int segment_samples = *)



# XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
# XXX          Segment Hierarchical Pitman-Yor Tree Hyperparameters          XXX
# XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

cdef void update_shpylm_d_theta(NPYLM npylm)
cdef void __recursive_shpylm_d_theta_preparation(SHPYLMNode node, np.ndarray sum1_minus_y_ui, 
                                                np.ndarray sum1_minus_z_uwkj, np.ndarray sumy_ui,
                                                np.ndarray sumlogx_u)
cdef float __get_shpylm_1_minus_y_ui(SHPYLMNode node)
cdef float __get_shpylm_1_minus_z_uwkj_sum(SHPYLMNode node)
cdef float __get_shpylm_y_ui_sum(SHPYLMNode node)
cdef float __get_shpylm_logx_u(SHPYLMNode node)




# XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
# XXX         Character Hierarchical Pitman-Yor Tree Hyperparameters        XXX
# XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

cdef void update_chpylm_d_theta(NPYLM npylm)
cdef void update_chpylm_d_theta(NPYLM npylm)
cdef void __recursive_chpylm_d_theta_preparation(CHPYLMNode node, np.ndarray sum1_minus_y_ui, 
                                                np.ndarray sum1_minus_z_uwkj, np.ndarray sumy_ui,
                                                np.ndarray sumlogx_u)
cdef float __get_chpylm_1_minus_y_ui(CHPYLMNode node)
cdef float __get_chpylm_1_minus_z_uwkj_sum(CHPYLMNode node)
cdef float __get_chpylm_y_ui_sum(CHPYLMNode node)
cdef float __get_chpylm_logx_u(CHPYLMNode node)


cdef str BOS = "<bos>" # beginning of sequence
cdef str EOS = "<eos>" # end of sequence
cdef str BOU = "<bou>" # beginning of unit
cdef str EOU = "<eou>" # end of unit

cdef class Sequence():
    def __cinit__(self, str sequence_string):
        """
        Constructor of Sequence data structure.

        Parameters
        ----------
        sequence_string : string
            sequence string in its very basic form (no prefixes/suffixes/..)
        """
        self.sequence_string = sequence_string
        self.segmentation = [sequence_string]


    cdef void set_segmentation(self, list borders):
        """
        Function that get list of borders of segmetns and it will reupload sequence's segmentation
        regarding those borders.

        example:
            borders=[0, 3, 4, 10], sequence_string="abcdefghij" -> segmentation=["abc", "d", "efghij"]

        Parameters
        ----------
        borders : list of ints
            list of indices of sequence_string string chars (in other word char array),
            that describe segment borders - they have to be order by ascending
        """
        cdef list segmented_sequence = []
        cdef int i, j
        for i, j in zip(borders[:-1], borders[1:]):
            segmented_sequence.append(self.sequence_string[i:j])
        self.segmentation = segmented_sequence


from nhpylm.sequence cimport Sequence, BOS, EOS, BOU, EOU
from nhpylm.random_utils cimport random_choice, poisson
import numpy as np
cimport numpy as np
from libc.math cimport log
np.import_array()
DTYPE = np.float64


cdef str EPSILON = "ε" # an empty character

cdef class NPYLM():
    def __cinit__(self, int max_segment_size, float init_d, float init_theta, 
                float init_a, float init_b, set character_vocabulary, 
                float beta_stops, float beta_passes, float d_a, float d_b,
                float theta_alpha, float theta_beta):
        """
        Constructor of Nested Pitman-Yor Language Model.
        Initialize Segment Hierarchical Pitman-Yor Language Model.
            - current max depth
            - d and theta values for each depth
        Initialize Tone Hierarchical Pitman-Yor Language Model.
            - current max depth
            - d and theta values for each depth
            - beta stop/pass coefficients
            - init poisson a,b values
        Initialize all hyperparameters and NPYLM settings.

        Parameters
        ----------
        max_segment_size : int
            maximum segment size that the model considers
        init_d : float
            initial discount factor in G ~ PY(Gn, d, theta) - used in probability recursive computation
            it's used as a init value for each depth of both, chpylm and shpylm
            which is updated once the depth's values are learned
        init_theta : float
            initial theta that controls the average similarity of G and Gn in G ~ PY(Gn, d, theta) - used in probability recursive computation
            it's used as a init value for each depth of both, chpylm and shpylm
            which is updated once the depth's values are learned
        init_a : float
            initial 'a' argument in gamma distribution G(a, b) in order to estimate lambda for poisson correction
        init_b : float
            initial 'b' argument in gamma distribution G(a, b) in order to estimate lambda for poisson correction
        character_vocabulary : set
            character vocabulary set of all unique characters
        beta_stops : float
            stops hyperparameter, that will be used during sampling and probability computation in CHPYLM
            that helps us avoid the situation of zero stops in depth that we want to include
        beta_passes : float
            stops hyperparameter, that will be used during sampling and probability computation in CHPYLM
            that helps us to avoid the situation of zero passes in depth that we want to include
        d_a : float
            hyperparameter for the 'd' prediction, part of the alpha value of Beta prior
        d_b : float
            hyperparameter for the 'd' prediction, part of the beta value of Beta prior
        theta_alpha : float
            hyperparameter for the 'theta' prediction, part of the shape value of Gamma prior
        theta_beta : float
            hyperparameter for the 'theta' prediction, part of the scale value of Gamma prior
        """
        self.max_segment_size = max_segment_size
        self.init_d = init_d
        self.init_theta = init_theta
        self.d_a = d_a
        self.d_b = d_b
        self.theta_alpha = theta_alpha
        self.theta_beta = theta_beta
        self.character_vocabulary = character_vocabulary
        self.root_tables_context_lengths = {}
        # Init SHPYLM settings
        self.shpylm_max_depth = -1
        self.shpylm_ds = []
        self.shpylm_thetas = []
        self.shpylm_root = SHPYLMNode(EPSILON, None, self, 0)
        # Init CHPYLM settings
        self.chpylm_max_depth = -1
        self.chpylm_ds = []
        self.chpylm_thetas = []
        self.poisson_k_probs = [1.0/(max_segment_size+2) for _ in range(max_segment_size)]
        self.poisson_lambda = np.random.gamma(init_a, 1/init_b)
        self.beta_stops = beta_stops
        self.beta_passes = beta_passes
        self.init_poisson_a = init_a
        self.init_poisson_b = init_b
        self.chpylm_root = CHPYLMNode(EPSILON, None, self, 0)

    cdef void add_sequence(self, Sequence sequence):
        """
        Add sequence segments to NPYLM.
        Add all segments from sequence into SHPYLM with the bigram (one previous segment) context.
        In case of the newly created table in the root level of SHPYLM, add the segment into 
        CHPYLM.

        Parameters
        ----------
        sequence : Sequence
            sequence that contains actual segmentation we want to add to NPYLM
        """
        # First, add sequence into segment HPYLM
        cdef str first_gram, second_gram
        cdef bint added_table
        cdef float pwhcomma

        pwhcomma = self.get_G0_probability(sequence.segmentation[0])
        added_table = self.shpylm_root.add_segment(sequence.segmentation[0], [BOS], pwhcomma) # First bigram
        if added_table:
            self.add_segment_to_chpylm(sequence.segmentation[0])

        for first_gram, second_gram in zip(sequence.segmentation[:-1], sequence.segmentation[1:]): # All middle bigrams
            pwhcomma = self.get_G0_probability(second_gram)
            added_table = self.shpylm_root.add_segment(second_gram, [first_gram], pwhcomma)
            if added_table:
                self.add_segment_to_chpylm(second_gram)

        pwhcomma = self.get_G0_probability(EOS)
        added_table = self.shpylm_root.add_segment(EOS, [sequence.segmentation[-1]], pwhcomma) # Last bigram
        if added_table:
            self.chpylm_root.add_character(EOS, [], 1.0/len(self.character_vocabulary))

    cdef void remove_sequence(self, Sequence sequence):
        """
        Remove sequence segments from NPYLM.
        Remove all segments from sequence from SHPYLM with the bigram (one previous segment) context.
        In case of removed table in the root level of SHPYLM, remove also the segment from CHPYLM.

        Parameters
        ----------
        sequence : Sequence
            sequence that contains actual segmentation we want to remove from NPYLM
        """
        # First, remove sequence from segment HPYLM
        cdef str first_gram, second_gram
        cdef bint removed_table

        removed_table = self.shpylm_root.remove_segment(sequence.segmentation[0], [BOS]) # First bigram
        if removed_table:
            self.remove_segment_from_chpylm(sequence.segmentation[0])

        for first_gram, second_gram in zip(sequence.segmentation[:-1], sequence.segmentation[1:]):
            removed_table = self.shpylm_root.remove_segment(second_gram, [first_gram])
            if removed_table:
                self.remove_segment_from_chpylm(second_gram)

        removed_table = self.shpylm_root.remove_segment(EOS, [sequence.segmentation[-1]]) # Last bigram
        if removed_table:
            self.chpylm_root.remove_character(EOS, [])


    cdef void add_segment_to_chpylm(self, str segment):
        """
        Add segment to CHPYLM.
        Sample context length for all segment characters and add the each character with its sampled context length
        (number of previous characters, we are sampling n for character n-gram) into CHPYLM.
        All generated context lengths are stored at the same table index as index of the new table in SHPYLM 
        tables corresponding to the segment, that was created - which was the reason of invoking this method.

        Parameters
        ----------
        segment : str
            segment as a string, each character or special sign needs to be stored as a single character
            (!!<bos> or <eos> are always 5 special signs/characters long segments, not one!!)
        """
        cdef list segment_characters = [BOU]+[*segment]+[EOU]
        cdef int i, ngram
        cdef list context_lengths = []
        cdef CHPYLMNode chpylm_root = self.chpylm_root
        for i in range(0, len(segment_characters)):
            ngram = self.chpylm_root.sample_context_length(segment_characters[i], segment_characters[:i], 
                                                            1.0/len(self.character_vocabulary))
            context_lengths.append(ngram)
            chpylm_root.add_character(segment_characters[i], segment_characters[i-ngram:i], 1.0/len(self.character_vocabulary))

        # Store the mapping of sampled context lengths and tables in SHPYLM
        if segment in self.root_tables_context_lengths:
            self.root_tables_context_lengths[segment].append(context_lengths)
        else:
            self.root_tables_context_lengths[segment] = [context_lengths]


    cdef void remove_segment_from_chpylm(self, str segment):
        """
        Remove segment from CHPYLM.
        Remove all segment characters from CHPYLM using the context lengths information that was stored based on the
        last table index that was removed. The root_tables_context_lengths[segment] was generated in the same order
        as the SHPYLM.tables[segment].tables, so indices should match. Therefore use the same context lengths to 
        remove the character with the specific context from CHPYLM (from the specific tree depth).
        Finally, remove the context lengths from root_tables_context_lengths[segment] since the corresponding table
        doesn't exist anymore, to keep the same order of both arrays in the same way.

        Parameters
        ----------
        segment : str
            segment as a string, each character or special sign needs to be stored as a single character
            (!!<bos> or <eos> are always 5 special signs/characters long segments, not one!!)
        """
        cdef list segment_characters = [BOU]+[*segment]+[EOU]
        cdef int i, ngram
        cdef int k_index = self.last_removed_shpylm_table_index
        for i, ngram in zip(range(0, len(segment_characters)), self.root_tables_context_lengths[segment][k_index]):
            self.chpylm_root.remove_character(segment_characters[i], segment_characters[i-ngram:i])
        
        # Remove the context length set from the mapping
        del self.root_tables_context_lengths[segment][k_index]
        



    cdef float get_G0_probability(self, str segment):
        """
        When segment is EOS, just return 1/|V|.
        Otherwise compute G0 probability of given segment.
        The result probability is a product of all character probabilities based on the
        full previous context that is provided in comming 'segment' string.
        As a init p(w|h') is used the 1/V, where V is the size of character vocabulary.

        In case that segment is in the correct size range (based on provided max_segment_size),
        - which tells us that we are not in the "init chpylm" stage - apply poisson correction
        p(c1...ck) = (p(c1...ck, k | TPYLM)/p(k|TPYLM)) * Po(k, lambda)
        where k is the length of the segment and p(k|TPYLM) tells us a probability of segment to be
        size of k. Generally we would like to keep segments arround the average segment length, to avoid
        the one character segments that should be more probable.

        Parameters
        ----------
        segment : str
            segment as a string, each character or special sign needs to be stored as a single character
        Returns
        -------
        prob : float
            G0 probability, the probability with a lower context than the SHPYLM root - probability of
            segment 's' given the SHPYLM tree
        """
        if segment == EOS:
            return 1.0/len(self.character_vocabulary)
        # First, get prob from character HPYLM
        cdef float prob = 1.0
        cdef list segment_characters = [BOU] + [*segment] + [EOU]
        for i in range(1, len(segment_characters)):
            prob *= self.chpylm_root.get_pwh_probability(segment_characters[i], segment_characters[:i], 1.0/len(self.character_vocabulary), 1.0)

        # Second, apply Poisson correction
        cdef int k = len(segment)
        if k <= self.max_segment_size: # apply poisson only in case of segments to be in the correct size
            prob = (prob/self.poisson_k_probs[k-1])*poisson(k,self.poisson_lambda)
        return prob

    cdef float get_bigram_probability(self, str first_gram, str second_gram):
        """
        Compute bigram probability p(second_gram | first_gram) regarding the Pitman Yor Language Model.
        First compute G0, then use it to feed the SHPYLM to compute probability from G1 and G2.  

        Parameters
        ----------
        first_gram : string
            first segment as a string, first gram
        second_gram : string
            second segment as a string when the first gram is given
        Returns
        -------
        prob : float
            bigram probability p(second_gram | first_gram)
        """
        cdef float prob = self.get_G0_probability(second_gram)
        
        # Get prob from segment HPYLM
        return self.shpylm_root.get_pwh_probability(second_gram, [first_gram], prob)
    

    cdef float get_segmentation_log_probability(self, list sequence_segmentation):
        """
        Compute probability of the given sequence segmentation (product of all bigrams).
        Apply logarithm on it.
        And return it back.

        Parameters
        ----------
        sequence_segmentation : list of strings
            list of segments of sequence we want to compute its log probability
        Returns
        -------
        prob_log : float
            logarithm probability of sequence segmentation given NPYLM
        """
        cdef str first_gram, second_gram
        cdef float prob_log = 0.0
        prob_log += log(self.get_bigram_probability(BOS, sequence_segmentation[0]))
        for first_gram, second_gram in zip(sequence_segmentation[:-1], sequence_segmentation[1:]):
            prob_log += log(self.get_bigram_probability(first_gram, second_gram))
        prob_log += log(self.get_bigram_probability(sequence_segmentation[-1], EOS))
        return prob_log












# XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
# XXX          Segment Hierarchical Pitman-Yor Tree Node          XXX
# XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX





"""
Segment Hierarchilcal Pitman-Yor Language Model Node

SHPYLM is represented as a tree, starting with in the root with parent = null
Each node has information about actual 
    context ... prvevious segments for instance "aaa"(root) -> "bbb"(1. depth) -> "ccc"(2.depth), 
        segments are in the sequence stored as "ccc bbb aaa segment" (=4gram example p(segment | ccc bbb aaa)).
    tables ... dictionary of restaurants - list of tables for the specific segment (stored in STables data structure)
    c_wh ... dictionary of customer counts - sum of all custmer along all tables for the specific segment
    t_h ... number of all tables over all segments, only in this node (not considering parents/children)
    c_h ... sum of all numbers in c_wh over all segments
"""
cdef class SHPYLMNode():
    def __cinit__(self, str context, SHPYLMNode parent, NPYLM npylm, int depth):
        """
        Initialize the SHPYLM node.
        Extend lists of npylm d and theta hyperparameters if needed.

        Parameters
        ----------
        context : str
            segment that is predecessor of parent's contexts which are predecessors of current segment
        parent : SHPYLMNode
            SHPYLM parent node, or null if we are in the root node
        npylm : NPYLM
            NPYLM object that has information of all current model settings.
        depth : int
            depth of shpylm tree, root is 0
        """
        self.parent = parent
        self.children = {}
        self.context = context
        self.tables = {}
        self.c_wh = {}
        self.c_h = 0
        self.t_h = 0
        self.npylm = npylm
        self.depth = depth
        if depth > npylm.shpylm_max_depth:
            npylm.shpylm_max_depth = depth
                    
        if depth >= len(npylm.shpylm_ds) and depth >= len(npylm.shpylm_thetas):
            npylm.shpylm_ds.append(npylm.init_d)
            npylm.shpylm_thetas.append(npylm.init_theta)
    
    cdef bint add_segment(self, str segment, list context, float pwhcomma):
        """
        1. Compute p(w|h) of this context depending on the previous pwhcomma
        p(w|h) = ((c(w|h)-d*t_hw)/(theta + c(h))) + ((theta + d*t_h)/(theta+c(h)))*p(w|h') 
        2. Propagate the p(w|h), segment we want to add and list of the not placed context
        3. If we are in the depth we want to (all context segments are placed), add customer to restaurant
            we need p(w|h) for that ... if new tables will be generated (based on the sampling), 
            In case that new table was generated in the children's node, we also add customer
            to the restaurant of this node (for the table on the level below).

        Parameters
        ----------
        segment : str
            segment we want to add into SHPYLM
        context : list of strings
            list of not placed context (previous) segments
            the root takes all previous context segments, its children takes only context list without the last
            segment (the one most closest to the current segment we are adding) - those closer are already stored
            in parent nodes
        pwhcomma : float
            p(w|h'), probability with the smaller context
        Returns
        -------
        new_table : bint
            boolean that says whether the new table was created in this node or not
        """
        cdef bint new_table = False
        cdef bint new_prev_table = True
        cdef STables segment_restaurant
        cdef SHPYLMNode context_child
        cdef float pwh
        cdef float d = self.npylm.shpylm_ds[self.depth]
        cdef float theta = self.npylm.shpylm_thetas[self.depth]

        # Compute p(w|h)
        if segment in self.tables:
            segment_restaurant = self.tables[segment]
            pwh = ((self.c_wh[segment] - d*len(segment_restaurant.tables))/(theta + self.c_h)) \
                        + (((theta + d*self.t_h)/(theta + self.c_h)) * pwhcomma)
        else:
            # For not existing character in tables, use only the second part of equation
            pwh = ((theta + d*self.t_h)/(theta + self.c_h)) * pwhcomma


        # Propagate segments to the next level
        if len(context) > 0:
            if not context[-1] in self.children:
                context_child = SHPYLMNode(context[-1], self, self.npylm, self.depth + 1)
                self.children[context[-1]] = context_child
            else:
                context_child =  self.children[context[-1]]
            new_prev_table = context_child.add_segment(segment, context[:-1], pwh)


        # If new table was generated in the lower level, add the customer here
        if new_prev_table:
            if segment in self.c_wh and segment in self.tables:
                self.c_wh[segment] += 1
                segment_restaurant = self.tables[segment]
            else:
                self.c_wh[segment] = 1
                segment_restaurant = STables(segment, self)
                self.tables[segment] = segment_restaurant
            self.c_h += 1
            new_table = segment_restaurant.add_customer(pwhcomma, self.t_h)
            if new_table:
                self.t_h += 1

        return new_table
        



    cdef bint remove_segment(self, str segment, list context):
        """
        1. Propagate segment we want to remove and list of the not searched context
        2. If we are in the depth we want to (all context segments are searched/used), remove customer 
            from segment's restaurant ... 
            if there is a table that got empty (we have to remove it based on the sampling), 
            in the children's node, we also remove customer from the restuarant of this node specific for the segment

        Parameters
        ----------
        segment : str
            segment we want to remove from SHPYLM
        context : list of strings
            list of not used context (previous) segments for finding the node where we need to remove our segment
        Returns
        -------
        removed_table : bint
            boolean that says whether there is a table which was removed in this node or not
        """
        cdef SHPYLMNode context_child
        cdef bint removed_table = False
        cdef bint removed_prev_table = True
        cdef STables segment_restaurant
        cdef float d = self.npylm.shpylm_ds[self.depth]

        # Propagate removing to the next level
        if len(context) > 0:
            context_child = self.children[context[-1]]
            removed_prev_table = context_child.remove_segment(segment, context[:-1])
            if context_child.t_h == 0:
                self.children.pop(context[-1])


        # Remove customer from table on this level
        if removed_prev_table:
            self.c_wh[segment] -= 1
            self.c_h -= 1
            segment_restaurant = self.tables[segment]
            removed_table = segment_restaurant.remove_customer()
            if self.c_wh[segment] == 0 or len(segment_restaurant.tables) == 0:
                self.c_wh.pop(segment)
                self.tables.pop(segment)
            if removed_table:
                self.t_h -= 1

        return removed_table


    cdef float get_pwh_probability(self, str segment, list context, float pwhcomma):
        """
        1. Compute p(w|h) of this context depending on the previous pwhcomma
        p(w|h) = ((c(w|h)-d*t_hw)/(theta + c(h))) + ((theta + d*t_h)/(theta+c(h)))*p(w|h') 
        2. Propagate the p(w|h) till the there is no more context we are interested in
            and return back the final probability value


        Parameters
        ----------
        segment : str
            segment we want to add into SHPYLM
        context : list of strings
            list of not searched/used context (previous) segments
        pwhcomma : float
            p(w|h'), probability with the smaller context
        Returns
        -------
        prob : float
            the p(w|h) probability 
        """
        cdef float prob = pwhcomma
        cdef SHPYLMNode context_child
        cdef STables seg_tables
        cdef float d = self.npylm.shpylm_ds[self.depth]
        cdef float theta = self.npylm.shpylm_thetas[self.depth]
        if segment in self.tables:
            seg_tables = self.tables[segment]
            prob = ((self.c_wh[segment] - d*len(seg_tables.tables))/(theta + self.c_h)) \
                        + (((theta + d*self.t_h)/(theta + self.c_h)) * pwhcomma)
        else:
            prob = ((theta + d*self.t_h)/(theta + self.c_h)) * pwhcomma
        if len(context) > 0 and context[-1] in self.children:
            context_child = self.children[context[-1]]
            prob = context_child.get_pwh_probability(segment, context[:-1], prob)
        return prob


"""
Restaurant for the specific segment - Segment Tables
There are tables and each table has several seats (!non zero!).
List tables contains numbers of seats at table of index k.
"""
cdef class STables():
    def __cinit__(self, str segment, SHPYLMNode hpylmnode):
        """
        Init STables - Segment Tables - Segment Restaurant

        Parameters
        ----------
        segment : str
            the restaurant is created for only this segment at specific node (based on context)
        hpylmnode : SHPYLMNode
            the SHPYLM Node that owns this restaurant
        """
        self.tables = []
        self.hpylmnode = hpylmnode
        self.segment = segment

    
    cdef bint add_customer(self, float pwhcomma, int t_h):
        """
        Add customer into restaurant.
        Use the p(w|h') and choose the table for the customer with prob distribution
        (c_hwk-d) for kth table (c_whk is a count of customers with segment w, context h, that are sitting at table k)
        (theta+(d*t_h))*p(w|h') for the creating new table

        Parameters
        ----------
        pwhcomma : float
            p(w|h'), probability of segment with shorter context
        t_h : int
            number of all tables over all restaurant in the w|h node
            (node described by context h and segment w)
        Returns
        -------
        new_table : bint
            boolean that says whether the new table was created or not
        """
        cdef int table_k
        cdef int c_hwk # customer count at table 'k'
        cdef list posibilities = []
        cdef bint new_table = False
        cdef float d = self.hpylmnode.npylm.shpylm_ds[self.hpylmnode.depth]
        cdef float theta = self.hpylmnode.npylm.shpylm_thetas[self.hpylmnode.depth]

        for c_hwk in self.tables:
            posibilities.append(max(0, c_hwk - d))
        if len(posibilities) > 0:
            posibilities.append((theta+(d*t_h))*pwhcomma)
            table_k = random_choice(posibilities)
        else:
            table_k = 0 # create new table in the next step

        if table_k == len(self.tables):
            self.tables.append(1)
            new_table = True
        else:
            self.tables[table_k] += 1

        return new_table
    
    cdef bint remove_customer(self):
        """
        Remove customer from restaurant.
        Choose the table with probability distribution
        (c_hwk) for kth table (c_whk is a count of customers with segment w, context h, that are sitting at table k)
        and remove the customer from it.

        If the table gets empty, remove it.
        And store the information of removed table index into NPYLM last_removed_shpylm_table_index that will be used 
        removing from CHPYLM

        Returns
        -------
        removed_table : bint
            boolean that says whether there is a removed table
        """
        cdef int table_k
        cdef int c_hwk # customer count at table 'k'
        cdef list posibilities = []
        cdef bint removed_table = False

        for c_hwk in self.tables:
            posibilities.append(c_hwk) # during the adding process, the minus discount factor is included
        table_k = random_choice(posibilities)
        self.hpylmnode.npylm.last_removed_shpylm_table_index = table_k

        self.tables[table_k] -= 1
        if self.tables[table_k] == 0:
            del self.tables[table_k]
            removed_table = True

        return removed_table













# XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
# XXX         Character Hierarchical Pitman-Yor Tree Node        XXX
# XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
cdef chpylm_trehshold = 1e-12







"""
Tone Hierarchilcal Pitman-Yor Language Model Node

CHPYLM is represented as a tree, starting with in the root with parent = null
Each node has information about current 
    context ... for backtracking prvevious characters the same way as in the CHPYLM
    tables ... dictionary of restaurants - list of tables for the specific characters (stored in STables data structure)
    c_wh ... dictionary of customer counts - sum of all custmer along all tables for the specific character
    t_h ... number of all tables over all characters, only in this node (not considering parents/children)
    c_h ... sum of all numbers in c_wh over all characters
"""
cdef class CHPYLMNode():
    def __cinit__(self, str context, CHPYLMNode parent, NPYLM npylm, int depth):
        """
        Initialize the CHPYLM node.
        Extend lists of npylm d and theta hyperparameters if needed.

        Parameters
        ----------
        context : str
            character that is predecessor of parent's contexts which are predecessors of current character
        parent : CHPYLMNode
            CHPYLM parent node, or null if we are in the root node
        npylm : NPYLM
            NPYLM object that has information of all current model settings.
        depth : int
            depth of chpylm tree, root is 0
        """
        self.parent = parent
        self.children = {}
        self.context = context
        self.tables = {}
        self.c_wh = {}
        self.c_h = 0
        self.t_h = 0
        self.npylm = npylm
        self.stops = 0
        self.passes = 0
        self.depth = depth
        if depth > npylm.chpylm_max_depth:
            npylm.chpylm_max_depth = depth
            
        if depth >= len(npylm.chpylm_ds) and depth >= len(npylm.chpylm_thetas):
            npylm.chpylm_ds.append(npylm.init_d)
            npylm.chpylm_thetas.append(npylm.init_theta)
            

    cdef bint add_character(self, str character, list context, float pwhcomma):
        """
        Compute p(w|h) = ((c(w|h)-d*t_hw)/(theta + c(h))) + ((theta + d*t_h)/(theta+c(h)))*p(w|h'),
        propagate p(w|h) to the deeper nodes with higher context, till the context list is not empty.
        Then add the customer to the node. If the new table was created, add the new customer also to the
        parent node (and recursivaly the same about parent's parent).

        Parameters
        ----------
        character : str
            character we want to add into CHPYLM
        context : list of strings
            list of not placed context (previous) characters
        pwhcomma : float
            p(w|h'), probability of character w with the smaller context
        Returns
        -------
        new_table : bint
            boolean that says whether the new table was created in this node or not
        """
        cdef bint new_table = False
        cdef bint new_prev_table = False
        cdef CTables character_restaurant
        cdef CHPYLMNode context_child
        cdef float pwh
        cdef CHPYLMNode parent
        cdef float d = self.npylm.chpylm_ds[self.depth]
        cdef float theta = self.npylm.chpylm_thetas[self.depth]

        # Compute p(w|h)
        if character in self.tables:
            character_restaurant = self.tables[character]
            pwh = ((self.c_wh[character] - d*len(character_restaurant.tables))/(theta + self.c_h)) \
                        + (((theta + d*self.t_h)/(theta + self.c_h)) * pwhcomma)
        else:
            # For not existing character in tables, use only the second part of equation
            pwh = ((theta + d*self.t_h)/(theta + self.c_h)) * pwhcomma


        # Propagate segments to the next level
        if len(context) > 0:
            if not context[-1] in self.children:
                context_child = CHPYLMNode(context[-1], self, self.npylm, self.depth+1)
                self.children[context[-1]] = context_child
            else:
                context_child =  self.children[context[-1]]
            new_prev_table = context_child.add_character(character, context[:-1], pwh)
        else:
            if character in self.c_wh and character in self.tables:
                self.c_wh[character] += 1
                character_restaurant = self.tables[character]
            else:
                self.c_wh[character] = 1
                character_restaurant = CTables(character, self)
                self.tables[character] = character_restaurant
            self.c_h += 1
            new_table = character_restaurant.add_customer(pwhcomma, self.t_h)
            if new_table:
                self.t_h += 1
            # update stops and passes
            self.stops += 1
            parent = self.parent
            while parent != None:
                parent.passes += 1
                parent = parent.parent


        # If new table was generated in the lower level, add the customer here
        if new_prev_table:
            if character in self.c_wh and character in self.tables:
                self.c_wh[character] += 1
                character_restaurant = self.tables[character]
            else:
                self.c_wh[character] = 1
                character_restaurant = CTables(character, self)
                self.tables[character] = character_restaurant
            self.c_h += 1
            new_table = character_restaurant.add_customer(pwhcomma, self.t_h)
            if new_table:
                self.t_h += 1

        return new_table
        



    cdef bint remove_character(self, str character, list context):
        """
        1. Propagate character we want to remove and list of the not used/searched context yet
        2. If we are in the depth we want to (all context characters are searched/used), remove customer 
            from character's restaurant ... 
            if there is a table that got empty (we have to remove it based on the sampling), 
            in the children's node, we also remove customer from the restuarant of this node specific for the character

        Parameters
        ----------
        character : str
            character we want to remove from CHPYLM
        context : list of strings
            list of not used context (previous) characters for finding the node where we need to remove our character
        Returns
        -------
        removed_table : bint
            boolean that says whether there is a table which was removed in this node or not
        """
        cdef CHPYLMNode context_child
        cdef bint removed_table = False
        cdef bint removed_prev_table = False
        cdef CTables character_restaurant

        # Propagate removing to the next level
        if len(context) > 0:
            context_child = self.children[context[-1]]
            removed_prev_table = context_child.remove_character(character, context[:-1])
            if context_child.t_h == 0:
                self.children.pop(context[-1])
        else:
            self.c_wh[character] -= 1
            self.c_h -= 1
            character_restaurant = self.tables[character]
            removed_table = character_restaurant.remove_customer()
            if self.c_wh[character] == 0 or len(character_restaurant.tables) == 0:
                self.c_wh.pop(character)
                self.tables.pop(character)
            if removed_table:
                self.t_h -= 1
            # update stops and passes
            self.stops -= 1
            parent = self.parent
            while parent != None:
                parent.passes -= 1
                parent = parent.parent

        # Remove customer from table on this level
        if removed_prev_table:
            self.c_wh[character] -= 1
            self.c_h -= 1
            character_restaurant = self.tables[character]
            removed_table = character_restaurant.remove_customer()
            if self.c_wh[character] == 0 or len(character_restaurant.tables) == 0:
                self.c_wh.pop(character)
                self.tables.pop(character)
            if removed_table:
                self.t_h -= 1

        return removed_table


    cdef float get_pwh_probability(self, str character, list context, float pwhcomma, float prev_pass_product):
        """
        1. Compute stop probability (that this node is the one that is more often final context character (final context length) than
        the middle context character) as 
        stop_prob = (node_stop_counts+beta_stops)/(node_stop_counts+beta_stops+node_pass_counts+beta_passes)
        2. Compute pass probability the same way as 
        pass_prob = (node_pass_counts+beta_passes)/(node_stop_counts+beta_stops+node_pass_counts+beta_passes)
        
        3. Compute p(w|h) of this context depending on the previous pwhcomma
        p(w|h) = ((c(w|h)-d*t_hw)/(theta + c(h))) + ((theta + d*t_h)/(theta+c(h)))*p(w|h')

        4. Propagate the p(w|h) and prev_pass_prob_product*pass_prob till the there is no more context we are interested in
            and return back the final probability value
        5. sum up the new deeper probability (with longer context) with the current probability as
            prob = longer_context_prob + current_context_prob*stop_prob*prev_pass_prob_product
            where prev_pass_prob_product is a product of all previous node's pass probabilities (where the current one is not 
            used in this product)
            and return the prob to the previous node

        We consider infinite ngrams, so there is a threshold to end up the process earlier. If there is no context anymore, we can 
        still consider those deeper ngrams as well.

        Parameters
        ----------
        character : str
            character we want to add into CHPYLM
        context : list of strings
            list of not searched/used context (previous) characters
        pwhcomma : float
            p(w|h'), probability with the smaller context
        prev_pass_product : float
            product of all previous node's pass probabilities
        Returns
        -------
        prob : float
            probability of character w with the context h
        """
        cdef float pass_prob
        cdef float stop_prob
        cdef float prob = pwhcomma
        cdef float next_prob
        cdef float pwh
        cdef CHPYLMNode context_child
        cdef CTables seg_tables
        cdef float new_stop_prob
        cdef float d = self.npylm.chpylm_ds[self.depth]
        cdef float theta = self.npylm.chpylm_thetas[self.depth]
        stop_prob = (self.stops + self.npylm.beta_stops)/\
                (self.stops + self.npylm.beta_stops + self.passes + self.npylm.beta_passes)
        pass_prob = (self.passes + self.npylm.beta_passes)/\
                (self.stops + self.npylm.beta_stops + self.passes + self.npylm.beta_passes)

        # If the threshold condition is not fulfiled, return 0.0
        if chpylm_trehshold < stop_prob * prev_pass_product:
            # Compute current p(w|h), then get the next p(w|h) with bigger context, 
            # then apply stop and pass probabilities into equation and send it to the
            # previous layer.
            if character in self.tables:
                seg_tables = self.tables[character]
                pwh = ((self.c_wh[character] - d*len(seg_tables.tables))/(theta + self.c_h)) \
                            + (((theta + d*self.t_h)/(theta + self.c_h)) * pwhcomma)
            else:
                # For not existing character in tables, use only the second part of equation
                pwh = ((theta + d*self.t_h)/(theta + self.c_h)) * pwhcomma
            if len(context) > 0 and context[-1] in self.children:
                context_child = self.children[context[-1]]
                next_prob = context_child.get_pwh_probability(character, context[:-1], pwh, prev_pass_product*pass_prob)
                prob = next_prob + pwh*stop_prob*prev_pass_product
            else:
                # In the case of not existing context in tree, use the Beta priors
                prob = 0
                new_stop_prob = (self.npylm.beta_stops/(self.npylm.beta_stops + self.npylm.beta_passes))
                while chpylm_trehshold < new_stop_prob * prev_pass_product:
                    prob += new_stop_prob * prev_pass_product * pwh
                    prev_pass_product *= (self.npylm.beta_passes/(self.npylm.beta_stops + self.npylm.beta_passes))    
        else:
            prob = 0
        return prob


    cdef int sample_context_length(self, str character, list context, float pwhcomma):
        """
        Sample length (in range of 0 and len(context)) of context that should be added into CHPYLM.

        Parameters
        ----------
        character : str
            character we want to add into CHPYLM
        context : list of strings
            list of not searched/used context (previous) characters
        pwhcomma : float
            p(w|h'), probability with the smaller context
        Returns
        -------
        sampled_length : int
            context length that was sampled based on conditional probability distribution for the specific character and its context
        """
        if len(context) == 0:
            return 0
        cdef list sample_prob_table = []
        self.__fill_sample_prob_table(sample_prob_table, character, context, pwhcomma, 1)
        return random_choice(sample_prob_table)


    cdef void __fill_sample_prob_table(self, list sample_prob_table, str character, list context, 
                                        float pwhcomma, float prev_pass_product):
        """
        Fill the sample_prob_table - table of probabilities of all possible context lenghts
        For the one context length compute the probability as
        1. Compute stop probability (that this node is the one that is more often final context character (final context length) than
        the middle context character) as 
        stop_prob = (node_stop_counts+beta_stops)/(node_stop_counts+beta_stops+node_pass_counts+beta_passes)
        2. Compute pass probability the same way as 
        pass_prob = (node_pass_counts+beta_passes)/(node_stop_counts+beta_stops+node_pass_counts+beta_passes)
        
        3. Compute p(w|h) of this context depending on the previous pwhcomma
        p(w|h) = ((c(w|h)-d*t_hw)/(theta + c(h))) + ((theta + d*t_h)/(theta+c(h)))*p(w|h')

        context_length_prob = pwh*stop_prob*prev_pass_product


        Parameters
        ----------
        sample_prob_table : list of floats
            list of probabilities for all possible context lengths that should be filled at the and of the recursive process
        character : str
            character we want to add into CHPYLM
        context : list of strings
            list of not searched/used context (previous) characters
        pwhcomma : float
            p(w|h'), probability with the smaller context
        prev_pass_product : float
            product of all previous node's pass probabilities
        """
        cdef float pass_prob
        cdef float stop_prob
        cdef float pwh
        cdef CHPYLMNode context_child
        cdef CTables seg_tables
        cdef float new_stop_prob
        cdef float d = self.npylm.chpylm_ds[self.depth]
        cdef float theta = self.npylm.chpylm_thetas[self.depth]

        stop_prob = (self.stops + self.npylm.beta_stops)/\
                (self.stops + self.npylm.beta_stops + self.passes + self.npylm.beta_passes)
        pass_prob = (self.passes + self.npylm.beta_passes)/\
                (self.stops + self.npylm.beta_stops + self.passes + self.npylm.beta_passes)
        # Compute current p(w|h), then get the next p(w|h) with bigger context, 
        # then apply stop and pass probabilities into equation and send it to the
        # previous layer.
        if character in self.tables:
            seg_tables = self.tables[character]
            pwh = ((self.c_wh[character] - d*len(seg_tables.tables))/(theta + self.c_h)) \
                        + (((theta + d*self.t_h)/(theta + self.c_h)) * pwhcomma)
        else:
            # For not existing character in tables, use only the second part of equation
            pwh = ((theta + d*self.t_h)/(theta + self.c_h)) * pwhcomma
        sample_prob_table.append(pwh*stop_prob*prev_pass_product)

        if len(context) > 0 and context[-1] in self.children:
            context_child = self.children[context[-1]]
            context_child.__fill_sample_prob_table(sample_prob_table, character, context[:-1], pwh, prev_pass_product*pass_prob)
        else:
            # In the case of not existing context in tree, use the Beta priors
            new_stop_prob = (self.npylm.beta_stops/(self.npylm.beta_stops + self.npylm.beta_passes))
            prev_pass_product *= pass_prob
            for _ in context:
                sample_prob_table.append(new_stop_prob * prev_pass_product * pwh)
                prev_pass_product *= (self.npylm.beta_passes/(self.npylm.beta_stops + self.npylm.beta_passes))    
    



"""
Restaurant for the specific character - Tone Tables 
(it differs a little bit from STables so it's better to keep them separated
due to avoid some unnecessary)
There are tables and each table has several seats (!non zero!).
List tables contains numbers of seats at table of index k.
"""
cdef class CTables():
    def __cinit__(self, str character, CHPYLMNode hpylmnode):
        """
        Init CTables - Tone Tables - Tone Restaurant

        Parameters
        ----------
        character : str
            the restaurant is created for only this character at specific node (based on context)
        hpylmnode : CHPYLMNode
            the CHPYLM Node that owns this restaurant
        """
        self.tables = []
        self.hpylmnode = hpylmnode
        self.character = character

    
    cdef bint add_customer(self, float pwhcomma, int t_h):
        """
        Add customer into restaurant.
        Use the p(w|h') and choose the table for the customer with prob distribution
        (c_hwk-d) for kth table (c_whk is a count of customers with character w, context h, that are sitting at table k)
        (theta+(d*t_h))*p(w|h') for the creating new table

        Parameters
        ----------
        pwhcomma : float
            p(w|h'), probability of character with shorter context
        t_h : int
            number of all tables over all restaurant in the w|h node
            (node described by context h and character w)
        Returns
        -------
        new_table : bint
            boolean that says whether the new table was created or not
        """
        cdef int table_k
        cdef int c_hwk # customer count at table 'k'
        cdef list posibilities = []
        cdef bint new_table = False
        cdef float d = self.hpylmnode.npylm.chpylm_ds[self.hpylmnode.depth]
        cdef float theta = self.hpylmnode.npylm.chpylm_thetas[self.hpylmnode.depth]

        for c_hwk in self.tables:
            posibilities.append(max(0, c_hwk - d))
        if len(posibilities) > 0:
            posibilities.append((theta+(d*t_h))*pwhcomma)
            table_k = random_choice(posibilities)
        else:
            table_k = 0 # create new table in the next step

        if table_k == len(self.tables):
            self.tables.append(1)
            new_table = True
        else:
            self.tables[table_k] += 1

        return new_table
    
    cdef bint remove_customer(self):
        """
        Remove customer from restaurant.
        Choose the table with probability distribution
        (c_hwk) for kth table (c_whk is a count of customers with character w, context h, that are sitting at table k)
        and remove the customer from it.

        If the table gets empty, remove it.

        Returns
        -------
        removed_table : bint
            boolean that says whether there is a removed table
        """
        cdef int table_k
        cdef int c_hwk # customer count at table 'k'
        cdef list posibilities = []
        cdef bint removed_table = False

        for c_hwk in self.tables:
            posibilities.append(c_hwk) # during the adding process, the minus discount factor is included
        table_k = random_choice(posibilities)

        self.tables[table_k] -= 1
        if self.tables[table_k] == 0:
            del self.tables[table_k]
            removed_table = True

        return removed_table


from nhpylm.viterbi_algorithm cimport viterbi_segment_sequences
from nhpylm.blocked_gibbs_sampler cimport blocked_gibbs_iteration
from nhpylm.hyperparameters cimport apply_hyperparameters_learning
from nhpylm.npylm cimport NPYLM
from nhpylm.sequence cimport Sequence
from libc.math cimport exp
cimport numpy as np
import numpy as np
import logging


cdef class NHPYLMModel:
    cdef int max_segment_size
    cdef int n_gram
    cdef float init_d
    cdef float init_theta
    cdef float init_a
    cdef float init_b
    cdef float beta_stops
    cdef float beta_passes
    cdef float d_a
    cdef float d_b
    cdef float theta_alpha
    cdef float theta_beta
    cdef dict train_statistics, dev_statistics
    cdef NPYLM npylm

    
    def __init__(self, max_segment_size = 7, n_gram = 2,
                init_d = 0.5, init_theta = 2.0,
                init_a = 6.0, init_b = 0.8333,
                beta_stops = 1.0, beta_passes = 1.0,
                d_a = 1.0, d_b = 1.0, theta_alpha = 1.0, theta_beta = 1.0):
        """
        Init Nested Hirearchical Pitman-Yor Language Model for Sequences
        
        Parameters
        ----------
        max_segment_size : int
            maximum segment size that the model considers
        n_gram : int
            the n gram of word hirearchical pitman your language model, for now we only support bigrams
        init_d : float
            initial discount factor in G ~ PY(G0, d, theta) - used in probability recursive computation
            hyperparameter d is learned during training
        init_theta : float
            initial theta that controls the average similarity of G and G0 in G ~ PY(G0, d, theta) - used in probability recrusive computation
            hyperparameter theta is learned during trainign
        init_a : float
            initial 'a' argument in gamma distribution G(a, b) in order to estimate lambda for poisson correction
        init_b : float
            initial 'b' argument in gamma distribution G(a, b) in order to estimate lambda for poisson correction
        beta_stops : float
            stops hyperparameter, that will be used during sampling and probability computation in CHPYLM
            that helps us avoid the situation of zero stops in depth that we want to include
        beta_passes : float
            stops hyperparameter, that will be used during sampling and probability computation in CHPYLM
            that helps us to avoid the situation of zero passes in depth that we want to include
        d_a : float
            a parameter for d learning as a base of alpha of Beta distribution
        d_b : float
            b parameter for d learning as a base of beta of Beta distribution
        theta_alpha : float
            alpha parameter for theta learning as a base of shape of Gamma distribution
        theta_beta : float
            beta parameter for theta learning as base of scale of Gamma distribution
        """
        self.max_segment_size = max_segment_size
        self.n_gram = n_gram
        if not n_gram == 2:
            raise NotImplementedError("For now, we support only bigrams, but {} gram was given.".format(n_gram))
        self.init_d = init_d
        self.init_theta = init_theta
        self.init_a = init_a
        self.init_b = init_b
        self.beta_stops = beta_stops
        self.beta_passes = beta_passes
        self.d_a = d_a
        self.d_b = d_b
        self.theta_alpha = theta_alpha
        self.theta_beta = theta_beta
        logging.info("The NHPYLM model was initialized with max segment size {}".format(max_segment_size))


    cpdef void train(self, list train_data, list dev_data, int epochs, bint d_theta_learning, 
                    bint poisson_learning, int print_each_nth_iteration = 5):
        """
        Perform the training process of the NHPYLM model. Each iteration print current statistics.
        Apply the hyperparameter learning after each iteration. Function parameters specify those learnings.

        Parameters
        ----------
        train_data : list of strings
            list of training sequences, each represented as a string
        dev_data : list of strings
            list of dev sequences, each represented as a string
        epochs : int
            number of training epochs
        d_theta_learning : boolean
            whether we want to apply d theta learning after each epoch or not
        poisson_learning : boolean
            whether we want to apply poisson learning or not
        print_each_nth_iteration : int
            print only iterations modulo print_each_nth_iteration
        """
        cdef int i
        cdef str sequence_str
        cdef list train_sequences = []
        cdef list dev_sequences = []
        cdef set train_character_vocabulary = set()
        cdef str character
        cdef list train_segments, dev_segments
        cdef float train_perplexity, dev_perplexity
        cdef Sequence sequence
        logging.info("NHPYLM train - {} train sequences, {} dev sequences.".format(len(train_data), len(dev_data)))

        # Prepare Sequences
        for sequence_str in train_data:
            train_sequences.append(Sequence(sequence_str))
            for character in sequence_str:
                train_character_vocabulary.add(character)
        for sequence_str in dev_data:
            dev_sequences.append(Sequence(sequence_str))

        # Initialize NPYLM and load all training sequences to it
        self.npylm = NPYLM(self.max_segment_size, self.init_d, self.init_theta, 
                            self.init_a, self.init_b,
                            train_character_vocabulary,
                            self.beta_stops, self.beta_passes,
                            self.d_a, self.d_b, self.theta_alpha, self.theta_beta)
        for sequence in train_sequences:
            self.npylm.add_sequence(sequence)

        # Training
        for i in range(epochs):
            blocked_gibbs_iteration(self.npylm, train_sequences)
            apply_hyperparameters_learning(self.npylm, train_sequences, d_theta_learning, poisson_learning)

            if (i+1)%print_each_nth_iteration == 0:
                train_segments, train_perplexity = self.predict_segments(train_data)
                dev_segments, dev_perplexity = self.predict_segments(dev_data)
                logging.info("\n{}. iteration\n\tTrain Perplexity: {}\n\tDev Perplexity: {}".format(i+1, train_perplexity, dev_perplexity)) # TODO REPLACE BY PRINT


    cpdef tuple predict_segments(self, list sequences):
        """
        Segment sequences using this trained model. Compute perplexity.

        Parameters
        ----------
        sequences : list of strings
            list of training sequences, each represented as a string
        Returns
        -------
        segmented_sequences : list of lists of strings
            list of sequences, each sequence is represented as list of strings
        perplexity : float
            compute perplexity of segmented sequences
        """
        cdef float perplexity
        cdef float prob_sum = 0.0
        cdef list sequence_segmentation
        cdef list segmented_sequences = viterbi_segment_sequences(self.npylm, sequences)
        for sequence_segmentation in segmented_sequences:
            prob_sum += self.npylm.get_segmentation_log_probability(sequence_segmentation)/len(sequence_segmentation)
        perplexity = exp(-prob_sum/len(segmented_sequences))
        return segmented_sequences, perplexity


cdef class NHPYLMClassesModel:
    cdef int max_segment_size
    cdef int n_gram
    cdef float init_d
    cdef float init_theta
    cdef float init_a
    cdef float init_b
    cdef float beta_stops
    cdef float beta_passes
    cdef float d_a
    cdef float d_b
    cdef float theta_alpha
    cdef float theta_beta
    cdef dict train_statistics, dev_statistics
    cdef dict npylm_classes

    
    def __init__(self, max_segment_size = 7, n_gram = 2,
                init_d = 0.5, init_theta = 2.0,
                init_a = 6.0, init_b = 0.8333,
                beta_stops = 1.0, beta_passes = 1.0,
                d_a = 1.0, d_b = 1.0, theta_alpha = 1.0, theta_beta = 1.0):
        """
        Init Nested Hirearchical Pitman-Yor Language Model for Gregorian Sequences
        
        Parameters
        ----------
        max_segment_size : int
            maximum segment size that the model considers
        n_gram : int
            the n gram of word hirearchical pitman your language model, for now we only support bigrams
        init_d : float
            initial discount factor in G ~ PY(G0, d, theta) - used in probability recursive computation
            hyperparameter d is learned during training
        init_theta : float
            initial theta that controls the average similarity of G and G0 in G ~ PY(G0, d, theta) - used in probability recrusive computation
            hyperparameter theta is learned during trainign
        init_a : float
            initial 'a' argument in gamma distribution G(a, b) in order to estimate lambda for poisson correction
        init_b : float
            initial 'b' argument in gamma distribution G(a, b) in order to estimate lambda for poisson correction
        beta_stops : float
            stops hyperparameter, that will be used during sampling and probability computation in CHPYLM
            that helps us avoid the situation of zero stops in depth that we want to include
        beta_passes : float
            stops hyperparameter, that will be used during sampling and probability computation in CHPYLM
            that helps us to avoid the situation of zero passes in depth that we want to include
        d_a : float
            a parameter for d learning as a base of alpha of Beta distribution
        d_b : float
            b parameter for d learning as a base of beta of Beta distribution
        theta_alpha : float
            alpha parameter for theta learning as a base of shape of Gamma distribution
        theta_beta : float
            beta parameter for theta learning as base of scale of Gamma distribution
        """
        self.max_segment_size = max_segment_size
        self.n_gram = n_gram
        if not n_gram == 2:
            raise NotImplementedError("For now, we support only bigrams, but {} gram was given.".format(n_gram))
        self.init_d = init_d
        self.init_theta = init_theta
        self.init_a = init_a
        self.init_b = init_b
        self.beta_stops = beta_stops
        self.beta_passes = beta_passes
        self.d_a = d_a
        self.d_b = d_b
        self.theta_alpha = theta_alpha
        self.theta_beta = theta_beta
        logging.info("The NHPYLM model was initialized with max segment size {}".format(max_segment_size))


    cpdef void train(self, list train_data, list dev_data, list train_labels, list dev_labels, 
                    int epochs, bint d_theta_learning, bint poisson_learning, int print_each_nth_iteration = 5):
        """
        Initialize eight NHPYLMs, one nhpylm model for each class. Divide data and train NHPYLMs
        Apply the hyperparameter learning after each iteration. Function parameters specify those learnings.

        Parameters
        ----------
        train_data : list of strings
            list of training sequences, each represented as a string
        dev_data : list of strings
            list of dev sequences, each represented as a string
        train_labels : list of strings
            list of train labels
        dev_labels : list of strings
            list of dev labels
        epochs : int
            number of training epochs
        d_theta_learning : boolean
            whether we want to apply d theta learning after each epoch or not
        poisson_learning : boolean
            whether we want to apply poisson learning or not
        print_each_nth_iteration : int
            print only iterations modulo print_each_nth_iteration
        """
        cdef int i
        cdef str sequence_str
        cdef list train_sequences = []
        cdef list dev_sequences = []
        cdef set train_character_vocabulary = set()
        cdef str character
        cdef list train_segments, dev_segments
        cdef float train_perplexity, dev_perplexity
        cdef Sequence sequence
        cdef str cl
        cdef set classes_set
        cdef list classes
        cdef float correct, train_accuracy, dev_accuracy
        cdef str gold_y, pred_y
        logging.info("NHPYLM train - {} train sequences, {} dev sequences.".format(len(train_data), len(dev_data)))

        # Prepare Sequences
        for sequence_str in train_data:
            train_sequences.append(Sequence(sequence_str))
            for character in sequence_str:
                train_character_vocabulary.add(character)
        for sequence_str in dev_data:
            dev_sequences.append(Sequence(sequence_str))

        # Initialize NPYLM and load all training sequences to it
        self.npylm_classes = {}
        classes_set = set()
        for cl in train_labels:
            classes_set.add(cl)
        classes = list(classes_set)
        for cl in classes:
            self.npylm_classes[cl] = NPYLM(self.max_segment_size, self.init_d, self.init_theta, 
                            self.init_a, self.init_b,
                            train_character_vocabulary,
                            self.beta_stops, self.beta_passes,
                            self.d_a, self.d_b, self.theta_alpha, self.theta_beta)
        cdef dict train_sequences_classes = {}
        cdef NPYLM npylm
        for sequence, cl in zip(train_sequences, train_labels):
            npylm = self.npylm_classes[cl]
            npylm.add_sequence(sequence)
            if not cl in train_sequences_classes:
                train_sequences_classes[cl] = []
            train_sequences_classes[cl].append(sequence)

        # Training
        for i in range(epochs):
            for cl in classes:
                blocked_gibbs_iteration(self.npylm_classes[cl], train_sequences_classes[cl])
                apply_hyperparameters_learning(self.npylm_classes[cl], train_sequences_classes[cl], d_theta_learning, poisson_learning)

            if (i+1)%print_each_nth_iteration == 0:
                train_segments, train_perplexity, train_prediction = self.predict_segments_classes(train_data)
                dev_segments, dev_perplexity, dev_prediction = self.predict_segments_classes(dev_data)
                correct = 0.0
                for gold_y, pred_y in zip(train_labels, train_prediction):
                    if gold_y == pred_y:
                        correct += 1.0
                train_accuracy = 100.0*(correct/len(train_labels))
                correct = 0.0
                for gold_y, pred_y in zip(dev_labels, dev_prediction):
                    if gold_y == pred_y:
                        correct += 1.0
                dev_accuracy = 100.0*(correct/len(dev_labels))
                logging.info("\n{}. iteration\n\tTrain Perplexity: {}\n\tTrain Accuracy: {}\n\tDev Perplexity: {}\n\tDev Accuracy: {}".format(i+1, train_perplexity, train_accuracy, dev_perplexity, dev_accuracy)) # TODO REPLACE BY PRINT


    cpdef tuple predict_segments_classes(self, list sequences):
        """
        Call the predict_classes function. Based on its values, compute perplexity.

        Parameters
        ----------
        sequences : list of strings
            list of training sequences, each represented as a string
        Returns
        -------
        segmented_sequences : list of lists of strings
            list of sequences, each sequence is represented as list of strings
        perplexity : float
            compute perplexity of segmented sequences
        prediction : list of strings
            list of predicted classes
        """
        cdef list prediction
        cdef list segmentations
        cdef list segmentation_log_probs
        prediction, segmentations, segmentation_log_probs = self.predict_classes(sequences)
        cdef list segmentation
        cdef float log_prob
        cdef float perplexity
        cdef float prob_sum = 0.0
        for segmentation, log_prob in zip(segmentations, segmentation_log_probs):
            prob_sum += (log_prob/len(segmentation))
        perplexity = exp(-prob_sum/len(sequences))

        return segmentations, perplexity, prediction


    cpdef tuple predict_segments(self, list sequences):
        """
        Call the predict_classes function. Based on its values, compute perplexity.

        Parameters
        ----------
        sequences : list of strings
            list of training sequences, each represented as a string
        Returns
        -------
        segmented_sequences : list of lists of strings
            list of sequences, each sequence is represented as list of strings
        perplexity : float
            compute perplexity of segmented sequences
        """
        cdef list segmentations
        cdef list segmentation_log_probs
        _, segmentations, segmentation_log_probs = self.predict_classes(sequences)
        cdef list segmentation
        cdef float log_prob
        cdef float perplexity
        cdef float prob_sum = 0.0
        for segmentation, log_prob in zip(segmentations, segmentation_log_probs):
            prob_sum += (log_prob/len(segmentation))
        perplexity = exp(-prob_sum/len(sequences))

        return segmentations, perplexity


    cpdef tuple predict_classes(self, list sequences):
        """
        Predict classes for comming sequences using Bayes rule. Consider all classes and their top segmentations via viterbi algorithm.
        Choose the best one. Compute probs of rest classes for the fixed segmentation. Take argmax of class for the probability.
        Take the best segmentaiton and sum of the prob segmentation over all classes, which is a probability of sequence segmentaion.

        Parameters
        ----------
        sequences : list of strings
            list of sequences represented as strings of notes
        Returns
        -------
        prediction : list of strings
            list of predicted classes
        segmentations : list of lists of strings
            list of segmentations represented as list of string segments
        segmentation_log_probs : list of floats
            list of log probabilities of chosen segmentaitons (sum of segmentation over all classes)
        """
        cdef str cl
        cdef str cl2
        cdef dict segmented_sequences_classes = {}
        # Predict segmentations for each class
        for cl in self.npylm_classes:
            segmented_sequences_classes[cl] = viterbi_segment_sequences(self.npylm_classes[cl], sequences)
        
        # Find the best classes and its segmentations
        cdef list sequence_segmentation
        cdef list best_classes = []
        cdef str best_class
        cdef float best_prob
        cdef float mode_log_probability
        cdef float prob
        cdef int i
        cdef NPYLM npylm
        for i in range(len(sequences)):
            best_class = ""
            best_prob = -float('inf')
            for cl in self.npylm_classes:
                mode_log_probability = -float('inf')
                npylm = self.npylm_classes[cl]
                for cl2 in self.npylm_classes:
                    sequence_segmentation = segmented_sequences_classes[cl2][i]
                    mode_log_probability = np.logaddexp(mode_log_probability, npylm.get_segmentation_log_probability(sequence_segmentation))
                if mode_log_probability > best_prob:
                    best_class = cl
                    best_prob = mode_log_probability
            best_classes.append(best_class)

        # Check that there is no better class for the segmentation
        # otherwise rewrite the best class
        cdef list new_best_classes = []
        cdef str new_best_class
        for i, best_class in enumerate(best_classes):
            sequence_segmentation = segmented_sequences_classes[best_class][i]
            new_best_class = ""
            best_prob = -float('inf')
            for cl in self.npylm_classes:
                npylm = self.npylm_classes[cl]
                prob = npylm.get_segmentation_log_probability(sequence_segmentation)
                if prob > best_prob:
                    new_best_class = cl
                    best_prob = prob
            new_best_classes.append(new_best_class)

        # Compute sum of prob of best segmentation over all classes considering logarithm
        cdef list prediction = []
        cdef list segmentations = []
        cdef list segmentation_log_probs = []
        cdef float prob_sum
        for i, best_class in enumerate(new_best_classes):
            prediction.append(best_class)
            sequence_segmentation = segmented_sequences_classes[best_class][i]
            segmentations.append(sequence_segmentation)
            prob_sum = -float('inf')
            for cl in self.npylm_classes:
                npylm = self.npylm_classes[cl]
                prob_sum = np.logaddexp(prob_sum, npylm.get_segmentation_log_probability(sequence_segmentation) + np.log(1/8))
            segmentation_log_probs.append(prob_sum)

        return prediction, segmentations, segmentation_log_probs
        



from setuptools import setup, Extension
from Cython.Build import cythonize
import numpy

extensions = [
    Extension("nhpylm.blocked_gibbs_sampler", ["nhpylm/blocked_gibbs_sampler.pyx"], include_dirs=[numpy.get_include()]),
    Extension("nhpylm.hyperparameters", ["nhpylm/hyperparameters.pyx"], include_dirs=[numpy.get_include()]),
    Extension("nhpylm.models", ["nhpylm/models.pyx"], include_dirs=[numpy.get_include()]),
    Extension("nhpylm.npylm", ["nhpylm/npylm.pyx"], include_dirs=[numpy.get_include()]),
    Extension("nhpylm.random_utils", ["nhpylm/random_utils.pyx"], include_dirs=[numpy.get_include()]),
    Extension("nhpylm.sequence", ["nhpylm/sequence.pyx"], include_dirs=[numpy.get_include()]),
    Extension("nhpylm.viterbi_algorithm", ["nhpylm/viterbi_algorithm.pyx"], include_dirs=[numpy.get_include()]),
]

setup(
    name="nhpylm",
    version="0.0.1",
    packages=["nhpylm"],
    ext_modules=cythonize(extensions, language_level="3"),
    install_requires=["numpy<2.0", "cython==3.0.12"],
    include_package_data=True,
    zip_safe=False,
)

